import asyncio
import os
from logging import getLogger

from dotenv import load_dotenv

load_dotenv()

logger = getLogger(__name__)


async def test_blaxel_config():
    """Test blaxel client configuration."""
    print("Testing blaxel client configuration...")
    from blaxel.core.client import client
    from blaxel.core.client.api.models import list_models

    models = await list_models.asyncio(client=client)
    assert models is not None
    assert len(models) > 0
    logger.info([model.metadata.name for model in models])
    print(f"✓ Found {len(models)} models")


async def test_blaxel_client_credentials():
    """Test blaxel client with client credentials authentication."""
    print("Testing blaxel client credentials authentication...")
    os.environ["BL_WORKSPACE"] = os.getenv("BL_WORKSPACE") or "test-workspace"
    os.environ["BL_CLIENT_CREDENTIALS"] = os.getenv("BL_CLIENT_CREDENTIALS") or "test-credentials"
    os.environ["BL_ENV"] = "dev"

    from blaxel.core.client import client
    from blaxel.core.client.api.models import list_models

    models = await list_models.asyncio(client=client)
    assert models is not None
    logger.info([model.metadata.name for model in models])
    print(f"✓ Client credentials auth successful, found {len(models)} models")


async def test_blaxel_api_key():
    """Test blaxel client with API key authentication."""
    print("Testing blaxel API key authentication...")
    os.environ["BL_WORKSPACE"] = os.getenv("BL_WORKSPACE") or "test-workspace"
    os.environ["BL_API_KEY"] = os.getenv("BL_API_KEY") or "test-api-key"
    os.environ["BL_ENV"] = "dev"

    from blaxel.core.client import client
    from blaxel.core.client.api.models import list_models

    models = await list_models.asyncio(client=client)
    assert models is not None
    logger.info([model.metadata.name for model in models])
    print(f"✓ API key auth successful, found {len(models)} models")


async def main():
    """Main function for standalone execution."""
    # You can only execute one test at a time, cause of autoload
    try:
        await test_blaxel_config()
        print("✅ Blaxel config test passed!")
    except Exception as e:
        print(f"❌ Config test failed: {e}")

    # Uncomment to test other authentication methods
    # try:
    #     await test_blaxel_client_credentials()
    #     print("✅ Client credentials test passed!")
    # except Exception as e:
    #     print(f"❌ Client credentials test failed: {e}")

    # try:
    #     await test_blaxel_api_key()
    #     print("✅ API key test passed!")
    # except Exception as e:
    #     print(f"❌ API key test failed: {e}")


if __name__ == "__main__":
    asyncio.run(main())



from typing import Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="PublicIp")


@_attrs_define
class PublicIp:
    """
    Attributes:
        description (Union[Unset, str]): Description of the region/location
        ipv_4_cidrs (Union[Unset, list[str]]): List of public ipv4 addresses
        ipv_6_cidrs (Union[Unset, list[str]]): List of public ipv6 addresses
    """

    description: Union[Unset, str] = UNSET
    ipv_4_cidrs: Union[Unset, list[str]] = UNSET
    ipv_6_cidrs: Union[Unset, list[str]] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        description = self.description

        ipv_4_cidrs: Union[Unset, list[str]] = UNSET
        if not isinstance(self.ipv_4_cidrs, Unset):
            ipv_4_cidrs = self.ipv_4_cidrs

        ipv_6_cidrs: Union[Unset, list[str]] = UNSET
        if not isinstance(self.ipv_6_cidrs, Unset):
            ipv_6_cidrs = self.ipv_6_cidrs

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if description is not UNSET:
            field_dict["description"] = description
        if ipv_4_cidrs is not UNSET:
            field_dict["ipv4Cidrs"] = ipv_4_cidrs
        if ipv_6_cidrs is not UNSET:
            field_dict["ipv6Cidrs"] = ipv_6_cidrs

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        if not src_dict:
            return None
        d = src_dict.copy()
        description = d.pop("description", UNSET)

        ipv_4_cidrs = cast(list[str], d.pop("ipv4Cidrs", UNSET))

        ipv_6_cidrs = cast(list[str], d.pop("ipv6Cidrs", UNSET))

        public_ip = cls(
            description=description,
            ipv_4_cidrs=ipv_4_cidrs,
            ipv_6_cidrs=ipv_6_cidrs,
        )

        public_ip.additional_properties = d
        return public_ip

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import logging
from typing import Any, Optional

from anthropic import AsyncAnthropic
from cohere import AsyncClientV2
from mistralai.sdk import Mistral
from pydantic_ai.models import Model
from pydantic_ai.models.anthropic import AnthropicModel
from pydantic_ai.models.cohere import CohereModel
from pydantic_ai.models.gemini import GeminiModel
from pydantic_ai.models.mistral import MistralModel
from pydantic_ai.models.openai import OpenAIModel
from pydantic_ai.providers.anthropic import AnthropicProvider
from pydantic_ai.providers.cohere import CohereProvider
from pydantic_ai.providers.mistral import MistralProvider
from pydantic_ai.providers.openai import OpenAIProvider

from blaxel.core import bl_model as bl_model_core
from blaxel.core import settings
from blaxel.core.client import client

from .custom.gemini import GoogleGLAProvider

logger = logging.getLogger(__name__)


class TokenRefreshingModel(Model):
    """Model wrapper that transparently refreshes tokens before each API call."""

    def __init__(self, model_config: dict):
        self.model_config = model_config
        self._cached_model = None
        self._cached_token = None

    def _create_model(self) -> Model:
        """Create the model instance with current token."""
        config = self.model_config
        type = config["type"]
        model = config["model"]
        url = config["url"]
        kwargs = config.get("kwargs", {})

        if type == "mistral":
            return MistralModel(
                model_name=model,
                provider=MistralProvider(
                    mistral_client=Mistral(
                        api_key=settings.auth.token,
                        server_url=url,
                    ),
                    **kwargs,
                ),
            )
        elif type == "cohere":
            return CohereModel(
                model_name=model,
                provider=CohereProvider(
                    cohere_client=AsyncClientV2(
                        api_key=settings.auth.token,
                        base_url=url,
                    ),
                ),
            )
        elif type == "xai":
            return OpenAIModel(
                model_name=model,
                provider=OpenAIProvider(
                    base_url=f"{url}/v1", api_key=settings.auth.token, **kwargs
                ),
            )
        elif type == "deepseek":
            return OpenAIModel(
                model_name=model,
                provider=OpenAIProvider(
                    base_url=f"{url}/v1", api_key=settings.auth.token, **kwargs
                ),
            )
        elif type == "cerebras":
            return OpenAIModel(
                model_name=model,
                provider=OpenAIProvider(
                    base_url=f"{url}/v1", api_key=settings.auth.token, **kwargs
                ),
            )
        elif type == "anthropic":
            return AnthropicModel(
                model_name=model,
                provider=AnthropicProvider(
                    anthropic_client=AsyncAnthropic(
                        api_key=settings.auth.token,
                        base_url=url,
                        default_headers=settings.auth.get_headers(),
                    ),
                    **kwargs,
                ),
            )
        elif type == "gemini":
            return GeminiModel(
                model_name=model,
                provider=GoogleGLAProvider(
                    api_key=settings.auth.token,
                    http_client=client.with_base_url(
                        f"{url}/v1beta/models"
                    ).get_async_httpx_client(),
                ),
            )
        else:
            if type != "openai":
                logger.warning(f"Model {model} is not supported by Pydantic, defaulting to OpenAI")
            return OpenAIModel(
                model_name=model,
                provider=OpenAIProvider(
                    base_url=f"{url}/v1", api_key=settings.auth.token, **kwargs
                ),
            )

    def _get_fresh_model(self) -> Model:
        """Get or create a model with fresh token if needed."""
        # Only refresh if using ClientCredentials (which has get_token method)
        if hasattr(settings.auth, "get_token"):
            # This will trigger token refresh if needed
            logger.debug(f"Calling get_token for {self.model_config['type']} model")
            settings.auth.get_token()

        new_token = settings.auth.token

        # If token changed or no cached model, create new one
        if self._cached_token != new_token or self._cached_model is None:
            self._cached_model = self._create_model()
            self._cached_token = new_token

        return self._cached_model

    @property
    def model_name(self) -> str:
        """Return the model name."""
        model = self._get_fresh_model()
        return model.model_name

    @property
    def system(self) -> Optional[Any]:
        """Return the system property from the wrapped model."""
        model = self._get_fresh_model()
        return model.system if hasattr(model, "system") else None

    async def request(self, *args, **kwargs):
        """Make a request to the model with token refresh."""
        model = self._get_fresh_model()
        # Pass all arguments through to the wrapped model
        return await model.request(*args, **kwargs)

    def __getattr__(self, name):
        """Delegate any other attributes to the wrapped model."""
        # Get fresh model in case token changed
        model = self._get_fresh_model()
        return getattr(model, name)


async def bl_model(name: str, **kwargs) -> Model:
    """Create a Pydantic AI model with automatic token refresh support."""
    url, type, model = await bl_model_core(name).get_parameters()

    # Store model configuration for recreation
    model_config = {"type": type, "model": model, "url": url, "kwargs": kwargs}

    # Create and return the wrapper
    return TokenRefreshingModel(model_config)



from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.custom_domain_spec_txt_records import CustomDomainSpecTxtRecords


T = TypeVar("T", bound="CustomDomainSpec")


@_attrs_define
class CustomDomainSpec:
    """Custom domain specification

    Attributes:
        cname_records (Union[Unset, str]): CNAME target for the domain
        last_verified_at (Union[Unset, str]): Last verification attempt timestamp
        status (Union[Unset, str]): Current status of the domain (pending, verified, failed)
        txt_records (Union[Unset, CustomDomainSpecTxtRecords]): Map of TXT record names to values for domain
            verification
        verification_error (Union[Unset, str]): Error message if verification failed
    """

    cname_records: Union[Unset, str] = UNSET
    last_verified_at: Union[Unset, str] = UNSET
    status: Union[Unset, str] = UNSET
    txt_records: Union[Unset, "CustomDomainSpecTxtRecords"] = UNSET
    verification_error: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        cname_records = self.cname_records

        last_verified_at = self.last_verified_at

        status = self.status

        txt_records: Union[Unset, dict[str, Any]] = UNSET
        if self.txt_records and not isinstance(self.txt_records, Unset) and not isinstance(self.txt_records, dict):
            txt_records = self.txt_records.to_dict()
        elif self.txt_records and isinstance(self.txt_records, dict):
            txt_records = self.txt_records

        verification_error = self.verification_error

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if cname_records is not UNSET:
            field_dict["cnameRecords"] = cname_records
        if last_verified_at is not UNSET:
            field_dict["lastVerifiedAt"] = last_verified_at
        if status is not UNSET:
            field_dict["status"] = status
        if txt_records is not UNSET:
            field_dict["txtRecords"] = txt_records
        if verification_error is not UNSET:
            field_dict["verificationError"] = verification_error

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.custom_domain_spec_txt_records import CustomDomainSpecTxtRecords

        if not src_dict:
            return None
        d = src_dict.copy()
        cname_records = d.pop("cnameRecords", UNSET)

        last_verified_at = d.pop("lastVerifiedAt", UNSET)

        status = d.pop("status", UNSET)

        _txt_records = d.pop("txtRecords", UNSET)
        txt_records: Union[Unset, CustomDomainSpecTxtRecords]
        if isinstance(_txt_records, Unset):
            txt_records = UNSET
        else:
            txt_records = CustomDomainSpecTxtRecords.from_dict(_txt_records)

        verification_error = d.pop("verificationError", UNSET)

        custom_domain_spec = cls(
            cname_records=cname_records,
            last_verified_at=last_verified_at,
            status=status,
            txt_records=txt_records,
            verification_error=verification_error,
        )

        custom_domain_spec.additional_properties = d
        return custom_domain_spec

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import asyncio

from blaxel.core import SandboxInstance


async def main():
    sandbox = await SandboxInstance.create()

    print(f"sandbox created => {sandbox.metadata.name}")

    result_expected = """Hello, world!
Hello, world!
Hello, world!
Hello, world!
Hello, world!
Hello, world!
"""

    # Start the long-running process
    await sandbox.process.exec(
        {
            "name": "test",
            "command": 'i=1; while [ $i -le 3 ]; do echo "Hello, world!"; sleep 31; echo "Hello, world!"; i=$((i+1)); done',
        }
    )

    result = ""

    # Define the callback for log handling
    def on_log(log: str):
        nonlocal result
        print(f"received log => {log}")
        result += log + "\n"

    # Use the auto-reconnecting wrapper
    stream_control = sandbox.process.stream_logs("test", {"on_log": on_log})

    # Wait for process to finish
    await sandbox.process.wait("test", max_wait=1000000)

    print(f"\nresult => {result}")
    print(f"resultExpected => {result_expected}")
    print(f"result === resultExpected => {result == result_expected}")

    if result != result_expected:
        raise Exception("Result does not match expected result")

    # Stop the streaming
    stream_control["close"]()

    await SandboxInstance.delete(sandbox.metadata.name)


if __name__ == "__main__":
    asyncio.run(main())



from typing import Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

T = TypeVar("T", bound="ProcessLogs")


@_attrs_define
class ProcessLogs:
    """
    Attributes:
        logs (str):  Example: logs output.
        stderr (str):  Example: stderr output.
        stdout (str):  Example: stdout output.
    """

    logs: str
    stderr: str
    stdout: str
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        logs = self.logs

        stderr = self.stderr

        stdout = self.stdout

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "logs": logs,
                "stderr": stderr,
                "stdout": stdout,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        if not src_dict:
            return None
        d = src_dict.copy()
        logs = d.pop("logs")

        stderr = d.pop("stderr")

        stdout = d.pop("stdout")

        process_logs = cls(
            logs=logs,
            stderr=stderr,
            stdout=stdout,
        )

        process_logs.additional_properties = d
        return process_logs

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from blaxel.core.mcp.server import FastMCP

mcp = FastMCP("My App")


# Add an addition tool
@mcp.tool()
def add(a: int, b: int) -> int:
    """Add two numbers"""
    return a + b


# Add a dynamic greeting resource
@mcp.resource("greeting://{name}")
def get_greeting(name: str) -> str:
    """Get a personalized greeting"""
    return f"Hello, {name}!"


def main():
    """Main function for standalone execution."""
    print("Testing MCP server creation...")

    # Test server creation
    assert mcp is not None
    assert mcp.name == "My App"
    print("✓ MCP server creation test passed")

    # Test tool registration
    result = add(1, 2)
    assert result == 3
    print("✓ MCP tool registration test passed")

    # Test resource registration
    result = get_greeting("World")
    assert result == "Hello, World!"
    print("✓ MCP resource registration test passed")

    print("✅ All MCP server tests passed!")

    print("🚀 Starting MCP WebSocket server on port 8080...")
    # Run the server
    mcp.run(transport="ws")


if __name__ == "__main__":
    main()



"""Pytest configuration and shared fixtures."""

import os
from unittest.mock import patch

import pytest


@pytest.fixture(scope="session", autouse=True)
def setup_test_environment():
    """Set up test environment variables."""
    # Set test environment variables
    test_env = {
        "BL_WORKSPACE": "test-workspace",
        "BL_TYPE": "test",
        "BL_NAME": "test-component",
        "BL_DEBUG_TELEMETRY": "false",
        "BL_ENABLE_OPENTELEMETRY": "false",
    }

    # Store original values
    original_env = {}
    for key, value in test_env.items():
        original_env[key] = os.environ.get(key)
        os.environ[key] = value

    yield

    # Restore original values
    for key, original_value in original_env.items():
        if original_value is None:
            os.environ.pop(key, None)
        else:
            os.environ[key] = original_value


@pytest.fixture
def mock_client():
    """Mock the Blaxel client."""
    with patch("blaxel.core.client.client") as mock:
        yield mock


@pytest.fixture
def mock_sandbox_client():
    """Mock the sandbox client."""
    with patch("blaxel.core.sandbox.client") as mock:
        yield mock


@pytest.fixture
def mock_websocket_client():
    """Mock the websocket client for MCP."""
    with patch("blaxel.core.mcp.websocket_client") as mock:
        yield mock


@pytest.fixture
def clean_environment():
    """Clean environment fixture that removes test-specific env vars."""
    env_vars_to_clean = [
        "BL_FUNCTION_TEST_URL",
        "BL_AGENT_TEST_URL",
        "BL_JOB_TEST_URL",
        "BL_SANDBOX_TEST_URL",
    ]

    # Store original values
    original_values = {}
    for var in env_vars_to_clean:
        original_values[var] = os.environ.get(var)
        if var in os.environ:
            del os.environ[var]

    yield

    # Restore original values
    for var, original_value in original_values.items():
        if original_value is not None:
            os.environ[var] = original_value


# Mark all tests as asyncio by default
pytestmark = pytest.mark.asyncio



# ruff: noqa: E402
import os
from logging import getLogger

from dotenv import load_dotenv
from livekit import agents
from livekit.agents import Agent, AgentSession, RoomInputOptions
from livekit.plugins import deepgram, elevenlabs, noise_cancellation, silero
from livekit.plugins.turn_detector.multilingual import MultilingualModel

from blaxel.livekit import bl_model, bl_tools

load_dotenv()

logger = getLogger(__name__)

MODEL = "gpt-4o-mini"
# MODEL = "claude-3-7-sonnet-20250219"
# MODEL = "xai-grok-beta"
# MODEL = "cohere-command-r-plus" # x -> tool call not supported
# MODEL = "gemini-2-0-flash"
# MODEL = "deepseek-chat"
# MODEL = "mistral-large-latest"


class Assistant(Agent):
    def __init__(self, tools=[]) -> None:
        super().__init__(
            instructions="""
                You are a helpful voice AI assistant.
            """,
            tools=tools,
        )


async def entrypoint(ctx: agents.JobContext):
    await ctx.connect()

    tools = await bl_tools(["blaxel-search"])
    model = await bl_model(MODEL)
    session = AgentSession(
        stt=deepgram.STT(model="nova-3", language="multi"),
        llm=model,
        tts=elevenlabs.TTS(voice_id="ODq5zmih8GrVes37Dizd", model="eleven_multilingual_v2"),
        vad=silero.VAD.load(),
        turn_detection=MultilingualModel(),
    )

    await session.start(
        room=ctx.room,
        agent=Assistant(tools=tools),
        room_input_options=RoomInputOptions(
            noise_cancellation=noise_cancellation.BVC(),
        ),
    )

    await session.generate_reply(instructions="Greet the user and offer your assistance.")


if __name__ == "__main__":
    logger.info("Starting assistant...")
    agents.cli.run_app(
        agents.WorkerOptions(entrypoint_fnc=entrypoint, port=os.getenv("BL_SERVER_PORT", 80))
    )



"""
This module provides the ClientCredentials class, which handles client credentials-based
authentication for Blaxel. It manages token refreshing and authentication flows using
client credentials and refresh tokens.
"""

import time
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Generator, Optional

import requests
from httpx import Request, Response

from .types import BlaxelAuth


@dataclass
class DeviceLoginFinalizeResponse:
    access_token: str
    expires_in: int
    refresh_token: str
    token_type: str


class ClientCredentials(BlaxelAuth):
    """
    ClientCredentials auth that authenticates requests using client credentials.
    """

    expires_at: Optional[datetime] = None

    def get_headers(self):
        """
        Retrieves the authentication headers after ensuring tokens are valid.

        Returns:
            dict: A dictionary of headers with Bearer token and workspace.

        Raises:
            Exception: If token refresh fails.
        """
        err = self.get_token()
        if err:
            raise err
        return {
            "X-Blaxel-Authorization": f"Bearer {self.credentials.access_token}",
            "X-Blaxel-Workspace": self.workspace_name,
        }

    def _request_token(self, remaining_retries: int = 3) -> Optional[Exception]:
        """
        Makes the token request with recursive retry logic.

        Args:
            remaining_retries (int): Number of retry attempts remaining.

        Returns:
            Optional[Exception]: An exception if refreshing fails after all retries, otherwise None.
        """
        try:
            headers = {
                "Authorization": f"Basic {self.credentials.client_credentials}",
                "Content-Type": "application/json",
            }
            body = {"grant_type": "client_credentials"}
            response = requests.post(f"{self.base_url}/oauth/token", headers=headers, json=body)
            response.raise_for_status()
            creds = response.json()
            self.credentials.access_token = creds["access_token"]
            self.credentials.refresh_token = creds["refresh_token"]
            self.credentials.expires_in = creds["expires_in"]
            self.expires_at = datetime.now() + timedelta(seconds=self.credentials.expires_in)
            return None
        except Exception as e:
            if remaining_retries > 0:
                time.sleep(1)
                return self._request_token(remaining_retries - 1)
            return e

    def get_token(self) -> Optional[Exception]:
        """
        Checks if the access token needs to be refreshed and performs the refresh if necessary.
        Uses recursive retry logic for up to 3 attempts.

        Returns:
            Optional[Exception]: An exception if refreshing fails after all retries, otherwise None.
        """
        if self.need_token():
            return self._request_token()
        return None

    def need_token(self):
        if not self.expires_at:
            return True
        return datetime.now() > self.expires_at - timedelta(minutes=10)

    def auth_flow(self, request: Request) -> Generator[Request, Response, None]:
        """
        Processes the authentication flow by ensuring tokens are valid and adding necessary headers.

        Parameters:
            request (Request): The HTTP request to authenticate.

        Yields:
            Request: The authenticated request.

        Raises:
            Exception: If token refresh fails.
        """
        self.get_token()
        request.headers["X-Blaxel-Authorization"] = f"Bearer {self.credentials.access_token}"
        request.headers["X-Blaxel-Workspace"] = self.workspace_name
        yield request

    @property
    def token(self):
        if not self.credentials.access_token:
            self.get_token()
        return self.credentials.access_token



import asyncio
import os
from logging import getLogger

from utils import create_or_get_sandbox

from blaxel.core.sandbox import SandboxInstance
from blaxel.core.sandbox.client.models.process_request import ProcessRequest

logger = getLogger(__name__)

SANDBOX_NAME = "sandbox-test-python-comprehensive"


async def test_filesystem(sandbox: SandboxInstance):
    """Test comprehensive filesystem operations."""
    print("🔧 Testing filesystem operations...")

    user = os.environ.get("USER", "testuser")
    test_file = f"/Users/{user}/Downloads/test"
    test_dir = f"/Users/{user}/Downloads/test2"

    # Test write and read
    await sandbox.fs.write(test_file, "Hello world")
    content = await sandbox.fs.read(test_file)
    assert content == "Hello world", f"File content mismatch: {content}"

    # Test directory listing
    dir_listing = await sandbox.fs.ls(f"/Users/{user}/Downloads")
    assert any(f.path == test_file for f in dir_listing.files), "File not found in directory"

    # Test mkdir
    await sandbox.fs.mkdir(test_dir)
    after_mkdir = await sandbox.fs.ls(test_dir)
    assert hasattr(after_mkdir, "files"), "Directory creation failed"

    # Test copy
    await sandbox.fs.cp(test_file, f"{test_dir}/test")
    after_cp = await sandbox.fs.ls(test_dir)
    assert any(f.path == f"{test_dir}/test" for f in after_cp.files), "File copy failed"

    # Test remove
    await sandbox.fs.rm(test_file)

    # Test recursive remove (should fail first, then succeed)
    try:
        await sandbox.fs.rm(test_dir)
        assert False, "Should have failed to remove non-empty directory"
    except Exception as e:
        logger.info(f"Expected error: {e}")

    await sandbox.fs.rm(test_dir, recursive=True)

    print("✅ Filesystem tests passed!")


async def test_process(sandbox: SandboxInstance):
    """Test process execution and management."""
    print("🔧 Testing process operations...")

    # Test process execution
    process_request = ProcessRequest(name="test", command="echo 'Hello world'")
    process = await sandbox.process.exec(process_request)

    assert process.status != "completed", "Process completed immediately"

    # Wait a bit for completion
    await asyncio.sleep(0.1)

    completed_process = await sandbox.process.get("test")
    if completed_process.status != "completed":
        # Wait for completion
        completed_process = await sandbox.process.wait("test")

    assert completed_process.status == "completed", "Process did not complete"

    # Test logs
    logs = await sandbox.process.logs("test")
    assert "Hello world" in logs, f"Logs incorrect: {logs}"

    # Test kill (should fail for completed process)
    try:
        await sandbox.process.kill("test")
        logger.info("Kill succeeded (process may have been running)")
    except Exception as e:
        logger.info(f"Expected error killing completed process: {e}")

    print("✅ Process tests passed!")


async def test_process_logs_streaming(sandbox: SandboxInstance):
    """Test process log streaming functionality."""
    print("🔧 Testing process log streaming...")

    log_called = False
    stdout_called = False
    stderr_called = False
    log_output = ""
    stdout_output = ""
    stderr_output = ""

    # Command that outputs to both stdout and stderr
    command = 'sh -c \'for i in $(seq 1 3); do echo "Hello from stdout $i"; echo "Hello from stderr $i" 1>&2; sleep 1; done\''
    name = "test-streaming"

    process_request = ProcessRequest(name=name, command=command)
    await sandbox.process.exec(process_request)

    def on_log(log):
        nonlocal log_called, log_output
        log_called = True
        log_output += log + "\n"
        print(f"onLog: {log}")

    def on_stdout(stdout):
        nonlocal stdout_called, stdout_output
        stdout_called = True
        stdout_output += stdout + "\n"
        print(f"onStdout: {stdout}")

    def on_stderr(stderr):
        nonlocal stderr_called, stderr_output
        stderr_called = True
        stderr_output += stderr + "\n"
        print(f"onStderr: {stderr}")

    stream = sandbox.process.stream_logs(
        name, {"on_log": on_log, "on_stdout": on_stdout, "on_stderr": on_stderr}
    )

    await sandbox.process.wait(name)
    stream["close"]()

    # Check that handlers were called
    assert log_called, "onLog was not called"
    assert stdout_called, "onStdout was not called"
    assert stderr_called, "onStderr was not called"
    assert "Hello from stdout" in log_output, f"Log output incorrect: {log_output}"
    assert "Hello from stdout" in stdout_output, f"Stdout output incorrect: {stdout_output}"
    assert "Hello from stderr" in stderr_output, f"Stderr output incorrect: {stderr_output}"

    print("✅ Process streaming tests passed!")


async def test_previews_public(sandbox: SandboxInstance):
    """Test public preview functionality."""
    print("🔧 Testing public previews...")

    try:
        await sandbox.previews.create(
            {
                "metadata": {"name": "preview-test-public"},
                "spec": {"port": 443, "prefix_url": "small-prefix-python", "public": True},
            }
        )

        previews = await sandbox.previews.list()
        assert len(previews) >= 1, "No previews found"

        retrieved_preview = await sandbox.previews.get("preview-test-public")
        assert retrieved_preview.name == "preview-test-public", "Preview name mismatch"

        url = retrieved_preview.spec.url
        assert url is not None, "Preview URL is None"

        print(f"Preview URL: {url}")

        # Test if preview is accessible (may fail in test environment)
        try:
            import httpx

            async with httpx.AsyncClient() as client:
                response = await client.get(f"{url}/health", timeout=10.0)
                if response.status_code == 200:
                    print("✅ Preview is healthy!")
                else:
                    print(f"⚠️ Preview returned status {response.status_code}")
        except Exception as e:
            print(f"⚠️ Could not test preview health: {e}")

    except Exception as e:
        print(f"❌ Preview test error: {e}")
    finally:
        try:
            await sandbox.previews.delete("preview-test-public")
        except:
            pass

    print("✅ Public preview tests completed!")


async def test_previews_private(sandbox: SandboxInstance):
    """Test private preview with token functionality."""
    print("🔧 Testing private previews with tokens...")

    try:
        # Create private preview with token
        await sandbox.previews.create(
            {
                "metadata": {"name": "preview-test-private"},
                "spec": {"port": 443, "prefix_url": "private-prefix-python", "public": False},
            }
        )

        # Create token for private preview
        token = await sandbox.previews.create_token("preview-test-private")
        print(f"Preview token created: {token.token[:20]}...")

        # Get preview with token
        private_preview = await sandbox.previews.get("preview-test-private")
        assert private_preview.name == "preview-test-private", "Private preview name mismatch"

        url = private_preview.spec.url
        print(f"Private Preview URL: {url}")

        # Test private preview access with token
        try:
            import httpx

            headers = {"Authorization": f"Bearer {token.token}"}
            async with httpx.AsyncClient() as client:
                response = await client.get(f"{url}/health", headers=headers, timeout=10.0)
                if response.status_code == 200:
                    print("✅ Private preview accessible with token!")
                else:
                    print(f"⚠️ Private preview returned status {response.status_code}")
        except Exception as e:
            print(f"⚠️ Could not test private preview health: {e}")

    except Exception as e:
        print(f"❌ Private preview test error: {e}")
    finally:
        try:
            await sandbox.previews.delete("preview-test-private")
        except:
            pass

    print("✅ Private preview tests completed!")


async def test_watch(sandbox: SandboxInstance):
    """Test filesystem watch functionality."""
    print("🔧 Testing filesystem watch...")

    watch_events = []

    def watch_callback(file_event):
        watch_events.append(file_event)
        print(f"Watch event: {file_event.op} {file_event.path}/{file_event.name}")

    def watch_callback_with_content(file_event):
        watch_events.append(file_event)
        print(
            f"Watch event with content: {file_event.op} {file_event.name} - {file_event.content[:50] if file_event.content else 'No content'}"
        )

    # Start watching
    handle = sandbox.fs.watch("/tmp", watch_callback)

    # Trigger some file operations
    await sandbox.fs.write("/tmp/watch-test.txt", "Initial content")
    await asyncio.sleep(1)

    await sandbox.fs.write("/tmp/watch-test.txt", "Modified content")
    await asyncio.sleep(1)

    await sandbox.fs.rm("/tmp/watch-test.txt")
    await asyncio.sleep(1)

    # Stop watching
    handle["close"]()

    print(f"Captured {len(watch_events)} watch events")
    for event in watch_events:
        print(f"  - {event.op}: {event.name}")

    print("✅ Watch tests completed!")


async def main():
    """Main comprehensive test function."""
    print("🚀 Starting comprehensive sandbox tests...")

    try:
        # Create or get sandbox
        sandbox = await create_or_get_sandbox(SANDBOX_NAME)
        print(f"✅ Sandbox ready: {sandbox.metadata.name}")

        # Run all test suites
        await test_filesystem(sandbox)
        print()

        await test_process(sandbox)
        print()

        await test_process_logs_streaming(sandbox)
        print()

        await test_previews_public(sandbox)
        print()

        await test_previews_private(sandbox)
        print()

        await test_watch(sandbox)
        print()

        print("🎉 All comprehensive tests completed successfully!")

    except Exception as e:
        print(f"❌ Comprehensive test failed with error: {e}")
        logger.exception("Comprehensive test error")
        raise
    finally:
        print("🧹 Final cleanup...")
        try:
            await SandboxInstance.delete(SANDBOX_NAME)
            print("✅ Sandbox deleted")
        except Exception as e:
            print(f"⚠️ Failed to delete sandbox: {e}")


if __name__ == "__main__":
    asyncio.run(main())



import asyncio
from typing import Any, Callable, Dict, Literal, Optional, Union

import httpx

from ..common.settings import settings
from .action import SandboxAction
from .client.models import ProcessResponse, SuccessResponse
from .client.models.process_request import ProcessRequest
from .types import ProcessRequestWithLog, ProcessResponseWithLog, SandboxConfiguration


class SandboxProcess(SandboxAction):
    def __init__(self, sandbox_config: SandboxConfiguration):
        super().__init__(sandbox_config)

    def stream_logs(
        self, process_name: str, options: Optional[Dict[str, Callable[[str], None]]] = None
    ) -> Dict[str, Callable[[], None]]:
        """Stream logs from a process with automatic reconnection and deduplication."""
        if options is None:
            options = {}

        reconnect_interval = 30  # 30 seconds in Python (TypeScript uses milliseconds)
        current_stream = None
        is_running = True
        reconnect_timer = None

        # Track seen logs to avoid duplicates
        seen_logs = set()

        def start_stream():
            nonlocal current_stream, reconnect_timer
            log_counter = [0]  # Use list to make it mutable in nested function

            # Close existing stream if any
            if current_stream:
                current_stream["close"]()

            # Create wrapper options with deduplication
            wrapped_options = {}

            if "on_log" in options:

                def deduplicated_on_log(log: str):
                    log_key = f"{log_counter[0]}:{log}"
                    log_counter[0] += 1
                    if log_key not in seen_logs:
                        seen_logs.add(log_key)
                        options["on_log"](log)

                wrapped_options["on_log"] = deduplicated_on_log

            if "on_stdout" in options:

                def deduplicated_on_stdout(stdout: str):
                    log_key = f"{log_counter[0]}:{stdout}"
                    log_counter[0] += 1
                    if log_key not in seen_logs:
                        seen_logs.add(log_key)
                        options["on_stdout"](stdout)

                wrapped_options["on_stdout"] = deduplicated_on_stdout

            if "on_stderr" in options:

                def deduplicated_on_stderr(stderr: str):
                    log_key = f"{log_counter[0]}:{stderr}"
                    log_counter[0] += 1
                    if log_key not in seen_logs:
                        seen_logs.add(log_key)
                        options["on_stderr"](stderr)

                wrapped_options["on_stderr"] = deduplicated_on_stderr

            # Start new stream with deduplication
            current_stream = self._stream_logs(process_name, wrapped_options)

            # Schedule next reconnection
            if is_running:

                def schedule_reconnect():
                    if is_running:
                        start_stream()

                reconnect_timer = asyncio.get_event_loop().call_later(
                    reconnect_interval, schedule_reconnect
                )

        # Start the initial stream
        start_stream()

        # Return control functions
        def close():
            nonlocal is_running, reconnect_timer, current_stream
            is_running = False

            # Cancel reconnect timer
            if reconnect_timer:
                reconnect_timer.cancel()
                reconnect_timer = None

            # Close current stream
            if current_stream:
                current_stream["close"]()
                current_stream = None

            # Clear seen logs
            seen_logs.clear()

        return {"close": close}

    def _stream_logs(
        self, identifier: str, options: Optional[Dict[str, Callable[[str], None]]] = None
    ) -> Dict[str, Callable[[], None]]:
        """Private method to stream logs from a process with callbacks for different output types."""
        if options is None:
            options = {}

        closed = False

        async def start_streaming():
            nonlocal closed

            url = f"{self.url}/process/{identifier}/logs/stream"
            headers = {**settings.headers, **self.sandbox_config.headers}

            try:
                async with httpx.AsyncClient() as client_instance:
                    async with client_instance.stream("GET", url, headers=headers) as response:
                        if response.status_code != 200:
                            raise Exception(f"Failed to stream logs: {await response.aread()}")

                        buffer = ""
                        async for chunk in response.aiter_text():
                            if closed:
                                break

                            buffer += chunk
                            lines = buffer.split("\n")
                            buffer = lines.pop()  # Keep incomplete line in buffer

                            for line in lines:
                                if line.startswith("stdout:"):
                                    content = line[7:]  # Remove 'stdout:' prefix
                                    if options.get("on_stdout"):
                                        options["on_stdout"](content)
                                    if options.get("on_log"):
                                        options["on_log"](content)
                                elif line.startswith("stderr:"):
                                    content = line[7:]  # Remove 'stderr:' prefix
                                    if options.get("on_stderr"):
                                        options["on_stderr"](content)
                                    if options.get("on_log"):
                                        options["on_log"](content)
                                else:
                                    if options.get("on_log"):
                                        options["on_log"](line)
            except Exception as e:
                # Suppress AbortError when closing
                if not (hasattr(e, "name") and e.name == "AbortError"):
                    raise e

        # Start streaming in the background
        task = asyncio.create_task(start_streaming())

        def close():
            nonlocal closed
            closed = True
            task.cancel()

        return {"close": close}

    async def exec(
        self, process: Union[ProcessRequest, ProcessRequestWithLog, Dict[str, Any]]
    ) -> Union[ProcessResponse, ProcessResponseWithLog]:
        """Execute a process in the sandbox."""
        on_log = None
        if isinstance(process, ProcessRequestWithLog):
            on_log = process.on_log
            process = process.to_dict()

        if isinstance(process, dict):
            if "on_log" in process:
                on_log = process["on_log"]
                del process["on_log"]
            process = ProcessRequest.from_dict(process)

        # Store original wait_for_completion setting
        should_wait_for_completion = process.wait_for_completion

        # Always start process without wait_for_completion to avoid server-side blocking
        if should_wait_for_completion and on_log is not None:
            process.wait_for_completion = False
        async with self.get_client() as client_instance:
            response = await client_instance.post("/process", json=process.to_dict())
            # Parse JSON response only once, with better error handling
            response_data = None
            if response.content:
                try:
                    response_data = response.json()
                except Exception:
                    # If JSON parsing fails, check the response first
                    self.handle_response_error(response)
                    raise

            self.handle_response_error(response)
            result = ProcessResponse.from_dict(response_data)

            # Handle wait_for_completion with parallel log streaming
            if should_wait_for_completion and on_log is not None:
                stream_control = self._stream_logs(result.pid, {"on_log": on_log})
                try:
                    # Wait for process completion
                    result = await self.wait(result.pid, interval=500, max_wait=1000 * 60 * 60)
                finally:
                    # Clean up log streaming
                    if stream_control:
                        stream_control["close"]()
            else:
                # For non-blocking execution, set up log streaming immediately if requested
                if on_log is not None:
                    stream_control = self._stream_logs(result.pid, {"on_log": on_log})
                    return ProcessResponseWithLog(
                        result, lambda: stream_control["close"]() if stream_control else None
                    )

            return result

    async def wait(
        self, identifier: str, max_wait: int = 60000, interval: int = 1000
    ) -> ProcessResponse:
        """Wait for a process to complete."""
        start_time = asyncio.get_event_loop().time() * 1000  # Convert to milliseconds
        status = "running"
        data = await self.get(identifier)

        while status == "running":
            await asyncio.sleep(interval / 1000)  # Convert to seconds
            try:
                data = await self.get(identifier)
                status = data.status or "running"
            except:
                break

            if (asyncio.get_event_loop().time() * 1000) - start_time > max_wait:
                raise Exception("Process did not finish in time")

        return data

    async def get(self, identifier: str) -> ProcessResponse:
        async with self.get_client() as client_instance:
            response = await client_instance.get(f"/process/{identifier}")
            self.handle_response_error(response)
            return ProcessResponse.from_dict(response.json())

    async def list(self) -> list[ProcessResponse]:
        async with self.get_client() as client_instance:
            response = await client_instance.get("/process")
            self.handle_response_error(response)
            return [ProcessResponse.from_dict(item) for item in response.json()]

    async def stop(self, identifier: str) -> SuccessResponse:
        async with self.get_client() as client_instance:
            response = await client_instance.delete(f"/process/{identifier}")
            self.handle_response_error(response)
            return SuccessResponse.from_dict(response.json())

    async def kill(self, identifier: str) -> SuccessResponse:
        async with self.get_client() as client_instance:
            response = await client_instance.delete(f"/process/{identifier}/kill")
            self.handle_response_error(response)
            return SuccessResponse.from_dict(response.json())

    async def logs(
        self, identifier: str, log_type: Literal["stdout", "stderr", "all"] = "all"
    ) -> str:
        async with self.get_client() as client_instance:
            response = await client_instance.get(f"/process/{identifier}/logs")
            self.handle_response_error(response)

            data = response.json()
            if log_type == "all":
                return data.get("logs", "")
            elif log_type == "stdout":
                return data.get("stdout", "")
            elif log_type == "stderr":
                return data.get("stderr", "")

            raise Exception("Unsupported log type")



from importlib.metadata import PackageNotFoundError
from importlib.metadata import version as package_version

from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry.instrumentation.llamaindex.utils import (
    _with_tracer_wrapper,
    process_request,
    process_response,
    start_as_current_span_async,
)
from opentelemetry.semconv_ai import SpanAttributes, TraceloopSpanKindValues
from opentelemetry.trace import get_tracer
from wrapt import wrap_function_wrapper

TO_INSTRUMENT = [
    {
        "class": "FunctionTool",
        "v9_module": "llama_index.tools.function_tool",
        "v10_module": "llama_index.core.tools.function_tool",
        "v10_legacy_module": "llama_index.legacy.tools.function_tool",
    },
    {
        "class": "QueryEngineTool",
        "v9_module": "llama_index.tools.query_engine",
        "v10_module": "llama_index.core.tools.query_engine",
        "v10_legacy_module": "llama_index.legacy.tools.query_engine",
    },
]


class BlaxelLlamaIndexInstrumentor(BaseInstrumentor):
    def instrumentation_dependencies(self):
        return ["llama_index"]

    def _instrument(self, **kwargs):
        tracer_provider = kwargs.get("tracer_provider")
        tracer = get_tracer(__name__, "0.2.0", tracer_provider)

        for module in TO_INSTRUMENT:
            try:
                package_version("llama-index-core")
                self._instrument_module(module["v10_module"], module["class"], tracer)
                self._instrument_module(module["v10_legacy_module"], module["class"], tracer)

            except PackageNotFoundError:
                self._instrument_module(module["v9_module"], module["class"], tracer)

    def _uninstrument(self, **kwargs):
        pass

    def _instrument_module(self, module_name, class_name, tracer):
        wrap_function_wrapper(module_name, f"{class_name}.call", query_wrapper(tracer))
        wrap_function_wrapper(module_name, f"{class_name}.acall", aquery_wrapper(tracer))


@_with_tracer_wrapper
def query_wrapper(tracer, wrapped, instance, args, kwargs):
    name = instance.__class__.__name__
    with tracer.start_as_current_span(f"{name}.tool") as span:
        span.set_attribute(
            SpanAttributes.TRACELOOP_SPAN_KIND,
            TraceloopSpanKindValues.TOOL.value,
        )
        span.set_attribute(SpanAttributes.TRACELOOP_ENTITY_NAME, name)

        process_request(span, args, kwargs)
        res = wrapped(*args, **kwargs)
        process_response(span, res)
        return res


@_with_tracer_wrapper
async def aquery_wrapper(tracer, wrapped, instance, args, kwargs):
    name = instance.__class__.__name__
    async with start_as_current_span_async(tracer=tracer, name=f"{name}.tool") as span:
        span.set_attribute(
            SpanAttributes.TRACELOOP_SPAN_KIND,
            TraceloopSpanKindValues.TOOL.value,
        )
        span.set_attribute(SpanAttributes.TRACELOOP_ENTITY_NAME, name)

        process_request(span, args, kwargs)
        res = await wrapped(*args, **kwargs)
        process_response(span, res)
        return res



import logging
import uuid
from typing import Any, Dict, List, Union

from ..client.api.compute.create_sandbox import asyncio as create_sandbox
from ..client.api.compute.delete_sandbox import asyncio as delete_sandbox
from ..client.api.compute.get_sandbox import asyncio as get_sandbox
from ..client.api.compute.list_sandboxes import asyncio as list_sandboxes
from ..client.api.compute.update_sandbox import asyncio as update_sandbox
from ..client.client import client
from ..client.models import Metadata, Runtime, Sandbox, SandboxSpec
from ..client.types import UNSET
from .filesystem import SandboxFileSystem
from .network import SandboxNetwork
from .preview import SandboxPreviews
from .process import SandboxProcess
from .session import SandboxSessions
from .types import (
    SandboxConfiguration,
    SandboxCreateConfiguration,
    SandboxUpdateMetadata,
    SessionWithToken,
)

logger = logging.getLogger(__name__)


class SandboxInstance:
    def __init__(self, sandbox: Sandbox):
        self.sandbox = sandbox
        self.config = SandboxConfiguration(sandbox)
        self.fs = SandboxFileSystem(self.config)
        self.process = SandboxProcess(self.config)
        self.previews = SandboxPreviews(sandbox)
        self.sessions = SandboxSessions(self.config)
        self.network = SandboxNetwork(self.config)

    @property
    def metadata(self):
        return self.sandbox.metadata

    @property
    def status(self):
        return self.sandbox.status

    @property
    def events(self):
        return self.sandbox.events

    @property
    def spec(self):
        return self.sandbox.spec

    async def wait(self, max_wait: int = 60000, interval: int = 1000) -> "SandboxInstance":
        logger.warning(
            "⚠️  Warning: sandbox.wait() is deprecated. You don't need to wait for the sandbox to be deployed anymore."
        )
        return self

    @classmethod
    async def create(
        cls,
        sandbox: Union[Sandbox, SandboxCreateConfiguration, Dict[str, Any], None] = None,
        safe: bool = True,
    ) -> "SandboxInstance":
        # Generate default values
        default_name = f"sandbox-{uuid.uuid4().hex[:8]}"
        default_image = "blaxel/prod-base:latest"
        default_memory = 4096

        # Handle SandboxCreateConfiguration or simple dict with name/image/memory/ports/envs keys
        if (
            sandbox is None
            or isinstance(sandbox, SandboxCreateConfiguration | dict)
            and (
                not isinstance(sandbox, Sandbox)
                and (
                    sandbox is None
                    or "name" in (sandbox if isinstance(sandbox, dict) else sandbox.__dict__)
                    or "image" in (sandbox if isinstance(sandbox, dict) else sandbox.__dict__)
                    or "memory" in (sandbox if isinstance(sandbox, dict) else sandbox.__dict__)
                    or "ports" in (sandbox if isinstance(sandbox, dict) else sandbox.__dict__)
                    or "envs" in (sandbox if isinstance(sandbox, dict) else sandbox.__dict__)
                    or "ttl" in (sandbox if isinstance(sandbox, dict) else sandbox.__dict__)
                    or "expires" in (sandbox if isinstance(sandbox, dict) else sandbox.__dict__)
                )
            )
        ):
            if sandbox is None:
                sandbox = SandboxCreateConfiguration()
            elif isinstance(sandbox, dict) and not isinstance(sandbox, Sandbox):
                sandbox = SandboxCreateConfiguration.from_dict(sandbox)

            # Set defaults if not provided
            name = sandbox.name or default_name
            image = sandbox.image or default_image
            memory = sandbox.memory or default_memory
            ports = sandbox._normalize_ports() or UNSET
            envs = sandbox._normalize_envs() or UNSET
            ttl = sandbox.ttl
            expires = sandbox.expires

            # Create full Sandbox object
            sandbox = Sandbox(
                metadata=Metadata(name=name),
                spec=SandboxSpec(
                    runtime=Runtime(
                        image=image, memory=memory, ports=ports, envs=envs, generation="mk3"
                    )
                ),
            )

            # Set ttl and expires if provided
            if ttl:
                sandbox.spec.runtime.ttl = ttl
            if expires:
                sandbox.spec.runtime.expires = expires.isoformat()
        else:
            # Handle existing Sandbox object or dict conversion
            if isinstance(sandbox, dict):
                sandbox = Sandbox.from_dict(sandbox)

            # Set defaults for missing fields
            if not sandbox.metadata:
                sandbox.metadata = Metadata(name=default_name)
            if not sandbox.spec:
                sandbox.spec = SandboxSpec(
                    runtime=Runtime(image=default_image, memory=default_memory)
                )
            if not sandbox.spec.runtime:
                sandbox.spec.runtime = Runtime(image=default_image, memory=default_memory)

            sandbox.spec.runtime.image = sandbox.spec.runtime.image or default_image
            sandbox.spec.runtime.memory = sandbox.spec.runtime.memory or default_memory
            sandbox.spec.runtime.generation = sandbox.spec.runtime.generation or "mk3"

        response = await create_sandbox(
            client=client,
            body=sandbox,
        )
        instance = cls(response)
        # TODO remove this part once we have a better way to handle this
        if safe:
            try:
                await instance.fs.ls("/")
            except Exception:
                pass
        return instance

    @classmethod
    async def get(cls, sandbox_name: str) -> "SandboxInstance":
        response = await get_sandbox(
            sandbox_name,
            client=client,
        )
        return cls(response)

    @classmethod
    async def list(cls) -> List["SandboxInstance"]:
        response = await list_sandboxes()
        return [cls(sandbox) for sandbox in response]

    @classmethod
    async def delete(cls, sandbox_name: str) -> Sandbox:
        response = await delete_sandbox(
            sandbox_name,
            client=client,
        )
        return response

    @classmethod
    async def update_metadata(
        cls, sandbox_name: str, metadata: SandboxUpdateMetadata
    ) -> "SandboxInstance":
        """Update sandbox metadata by merging new metadata with existing metadata.

        Args:
            sandbox_name: The name of the sandbox to update
            metadata: The metadata fields to update (labels and/or display_name)

        Returns:
            A new SandboxInstance with updated metadata
        """
        # Get the existing sandbox
        sandbox_instance = await cls.get(sandbox_name)
        sandbox = sandbox_instance.sandbox

        # Prepare the updated sandbox object
        updated_sandbox = Sandbox.from_dict(sandbox.to_dict())

        # Merge metadata
        if updated_sandbox.metadata is None:
            updated_sandbox.metadata = Metadata()

        # Update labels if provided
        if metadata.labels is not None:
            # Handle UNSET or None labels
            if updated_sandbox.metadata.labels is None or updated_sandbox.metadata.labels is UNSET:
                updated_sandbox.metadata.labels = {}
            else:
                # If labels exist, ensure it's a dict
                updated_sandbox.metadata.labels = dict(updated_sandbox.metadata.labels)
            updated_sandbox.metadata.labels.update(metadata.labels)

        # Update display_name if provided
        if metadata.display_name is not None:
            updated_sandbox.metadata.display_name = metadata.display_name

        # Call the update API
        response = await update_sandbox(
            sandbox_name=sandbox_name,
            client=client,
            body=updated_sandbox,
        )

        # Return new instance with updated sandbox
        return cls(response)

    @classmethod
    async def create_if_not_exists(
        cls, sandbox: Union[Sandbox, SandboxCreateConfiguration, Dict[str, Any]]
    ) -> "SandboxInstance":
        """Create a sandbox if it doesn't exist, otherwise return existing."""
        try:
            return await cls.create(sandbox)
        except Exception as e:
            # Check if it's a 409 conflict error (sandbox already exists)
            if (hasattr(e, "status_code") and e.status_code == 409) or (
                hasattr(e, "code") and e.code in [409, "SANDBOX_ALREADY_EXISTS"]
            ):
                # Extract name from different configuration types
                if isinstance(sandbox, SandboxCreateConfiguration):
                    name = sandbox.name
                elif isinstance(sandbox, dict):
                    if "name" in sandbox:
                        name = sandbox["name"]
                    elif "metadata" in sandbox and isinstance(sandbox["metadata"], dict):
                        name = sandbox["metadata"].get("name")
                    else:
                        name = None
                elif isinstance(sandbox, Sandbox):
                    name = sandbox.metadata.name if sandbox.metadata else None
                else:
                    name = None

                if not name:
                    raise ValueError("Sandbox name is required")

                sandbox_instance = await cls.get(name)
                return sandbox_instance
            raise e

    @classmethod
    async def from_session(
        cls, session: Union[SessionWithToken, Dict[str, Any]]
    ) -> "SandboxInstance":
        """Create a sandbox instance from a session with token."""
        if isinstance(session, dict):
            session = SessionWithToken.from_dict(session)

        # Create a minimal sandbox configuration for session-based access
        sandbox_name = session.name.split("-")[0] if "-" in session.name else session.name
        sandbox = Sandbox(metadata=Metadata(name=sandbox_name))
        config = SandboxConfiguration(
            sandbox=sandbox,
            force_url=session.url,
            headers={"X-Blaxel-Preview-Token": session.token},
            params={"bl_preview_token": session.token},
        )

        instance = cls.__new__(cls)
        instance.sandbox = sandbox
        instance.config = config
        instance.fs = SandboxFileSystem(config)
        instance.process = SandboxProcess(config)
        instance.previews = SandboxPreviews(sandbox)
        instance.sessions = SandboxSessions(config)
        instance.network = SandboxNetwork(config)

        return instance



# ruff: noqa: E402

import asyncio

from dotenv import load_dotenv

load_dotenv()

from logging import getLogger

from google.adk.agents import Agent
from google.adk.runners import Runner
from google.adk.sessions import InMemorySessionService
from google.genai import types

from blaxel.googleadk import bl_model, bl_tools

logger = getLogger(__name__)

MODEL = "gpt-4o-mini"
# MODEL = "claude-3-5-sonnet"
# MODEL = "xai-grok-beta"
# MODEL = "cohere-command-r-plus" # x -> tool call not supported
# MODEL = "gemini-2-0-flash"
# MODEL = "deepseek-chat"
# MODEL = "mistral-large-latest"


async def main():
    # Define constants for identifying the interaction context
    APP_NAME = "weather_agent"
    USER_ID = "user_1"
    SESSION_ID = "session_001"  # Using a fixed ID for simplicity

    tools = await bl_tools(["blaxel-search"], timeout_enabled=False)
    model = await bl_model(MODEL, temperature=0)

    description = "Provides weather information for specific cities."
    prompt = """You are a helpful weather assistant. Your primary goal is to provide current weather reports. "
When the user asks for the weather in a specific city,
Analyze the tool's response: if the status is 'error', inform the user politely about the error message.
If the status is 'success', present the weather 'report' clearly and concisely to the user.
Only use the tool when a city is mentioned for a weather request.
"""
    agent = Agent(
        model=model, name=APP_NAME, description=description, instruction=prompt, tools=tools
    )

    # Create the specific session where the conversation will happen
    session_service = InMemorySessionService()
    await session_service.create_session(app_name=APP_NAME, user_id=USER_ID, session_id=SESSION_ID)
    print(f"Session created: App='{APP_NAME}', User='{USER_ID}', Session='{SESSION_ID}'")

    runner = Runner(
        agent=agent,
        app_name=APP_NAME,
        session_service=session_service,
    )
    print(f"Runner created for agent '{runner.agent.name}'.")

    input = "What is the weather like in London?"
    content = types.Content(role="user", parts=[types.Part(text=input)])
    final_response_text = ""
    async for event in runner.run_async(
        user_id=USER_ID, session_id=SESSION_ID, new_message=content
    ):
        # Key Concept: is_final_response() marks the concluding message for the turn.
        if event.is_final_response():
            if event.content and event.content.parts:
                # Assuming text response in the first part
                final_response_text = event.content.parts[0].text
            elif event.actions and event.actions.escalate:  # Handle potential errors/escalations
                final_response_text = (
                    f"Agent escalated: {event.error_message or 'No specific message.'}"
                )

            # Add more checks here if needed (e.g., specific error codes)
            break  # Stop processing events once the final response is found
    print(f"Final response: {final_response_text}")


if __name__ == "__main__":
    asyncio.run(main())



from typing import Any, Dict, Optional, Type

from pydantic import BaseModel, Field, create_model

# Map JSON Schema types to Python types
json_type_mapping: Dict[str, Type] = {
    "string": str,
    "number": float,
    "integer": int,
    "boolean": bool,
    "object": dict,
    "array": list,
}

def create_model_from_json_schema(
    schema: Dict[str, Any], model_name: str = "DynamicModel"
) -> Type[BaseModel]:
    """
    To create a Pydantic model from the JSON Schema of MCP tools.

    Args:
        schema: A JSON Schema dictionary containing properties and required fields.
        model_name: The name of the model.

    Returns:
        A Pydantic model class.
    """
    properties = schema.get("properties", {})
    required_fields = set(schema.get("required", []))
    fields = {}

    for field_name, field_schema in properties.items():
        json_type = field_schema.get("type", "string")
        field_type = json_type_mapping.get(json_type, str)
        if field_name in required_fields:
            default_value = ...
        else:
            default_value = None
            field_type = Optional[field_type]
        fields[field_name] = (
            field_type,
            Field(default_value, description=field_schema.get("description", "")),
        )
    return create_model(model_name, **fields)


"""Tests for jobs functionality."""

from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from blaxel.core.jobs import bl_job


@pytest.mark.asyncio
async def test_bl_job_creation():
    """Test job creation."""
    job = bl_job("test-job")
    assert job.name == "test-job"
    assert str(job) == "Job test-job"


@pytest.mark.asyncio
async def test_bl_job_url_properties():
    """Test job URL properties."""
    job = bl_job("myjob")

    # Test that URL properties are accessible
    assert hasattr(job, "url")
    assert hasattr(job, "external_url")
    assert hasattr(job, "internal_url")
    assert hasattr(job, "forced_url")
    assert hasattr(job, "fallback_url")


@pytest.mark.asyncio
@patch("blaxel.core.jobs.client")
async def test_bl_job_run(mock_client):
    """Test job run functionality."""
    # Mock the HTTP client properly
    mock_httpx_client = MagicMock()
    mock_async_httpx_client = MagicMock()

    # Mock sync response
    mock_sync_response = MagicMock()
    mock_sync_response.status_code = 200
    mock_sync_response.text = "Job completed successfully"
    mock_httpx_client.post.return_value = mock_sync_response

    # Mock async response
    mock_async_response = MagicMock()
    mock_async_response.status_code = 200
    mock_async_response.text = "Job completed successfully"
    mock_async_httpx_client.post = AsyncMock(return_value=mock_async_response)

    mock_client.get_httpx_client.return_value = mock_httpx_client
    mock_client.get_async_httpx_client.return_value = mock_async_httpx_client

    job = bl_job("test-job")

    # Test sync run
    result = job.run([{"name": "charlou", "age": 25}])
    assert result == "Job completed successfully"

    # Test async run
    result = await job.arun([{"name": "charlou", "age": 25}])
    assert result == "Job completed successfully"


@pytest.mark.asyncio
async def test_bl_job_methods():
    """Test job has required methods."""
    job = bl_job("test-job")

    # Test that core methods exist
    assert hasattr(job, "run")
    assert hasattr(job, "arun")
    assert hasattr(job, "call")
    assert hasattr(job, "acall")


@pytest.mark.asyncio
async def test_bl_job_representation():
    """Test job string representation."""
    job = bl_job("test-job")

    # Test string representation
    assert str(job) == "Job test-job"
    assert repr(job) == "Job test-job"



import json
import logging

from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry.trace import Tracer, get_tracer

import blaxel.core.agents
import blaxel.core.jobs
from blaxel.telemetry.span import SpanManager

logger = logging.getLogger(__name__)


class BlaxelCoreInstrumentor(BaseInstrumentor):
    def instrumentation_dependencies(self):
        return []

    def _instrument(self, **kwargs):
        tracer_provider = kwargs.get("tracer_provider")
        tracer = get_tracer(__name__, "0.2.0", tracer_provider)
        self._patch_agent(tracer)
        self._patch_job(tracer)
        self._patch_tools(tracer)

    def _uninstrument(self, **kwargs):
        # Optionally restore original methods
        pass

    def _patch_agent(self, tracer: Tracer):
        orig_run = blaxel.core.agents.BlAgent.run

        def traced_run(self, *args, **kwargs):
            attributes = {
                "agent.name": self.name,
                "agent.args": str(args),
                "span.type": "agent.run",
                **SpanManager.get_default_attributes(),
            }
            with tracer.start_span(self.name, attributes=attributes) as span:
                try:
                    result = orig_run(self, *args, **kwargs)
                    span.set_attribute("agent.run.result", result)
                    return result
                except Exception as e:
                    span.set_attribute("agent.run.error", str(e))
                    raise

        blaxel.core.agents.BlAgent.run = traced_run

    def _patch_job(self, tracer: Tracer):
        orig_run = blaxel.core.jobs.BlJob.run

        def traced_run(self, *args, **kwargs):
            attributes = {
                "job.name": self.name,
                "span.type": "job.run",
                **SpanManager.get_default_attributes(),
            }
            with tracer.start_span(self.name, attributes=attributes) as span:
                try:
                    result = orig_run(self, *args, **kwargs)
                    span.set_attribute("job.run.result", result)
                    return result
                except Exception as e:
                    span.set_attribute("job.run.error", str(e))
                    raise

        blaxel.core.jobs.BlJob.run = traced_run

    def _patch_tools(self, tracer: Tracer):
        # Patch PersistentWebSocket.list_tools
        orig_list_tools = blaxel.core.tools.PersistentWebSocket.list_tools

        async def traced_list_tools(self, *args, **kwargs):
            span_attributes = {
                "tool.server": self._url,
                "tool.server_name": self.name,
                "span.type": "tool.list",
                **SpanManager.get_default_attributes(),
            }
            with tracer.start_span(self.name, attributes=span_attributes):
                result = await orig_list_tools(self, *args, **kwargs)
                # Optionally: span.set_attribute("tool.list.result", str(result))
                return result

        blaxel.core.tools.PersistentWebSocket.list_tools = traced_list_tools

        # Patch convert_mcp_tool_to_blaxel_tool to wrap tool calls
        orig_convert = blaxel.core.tools.convert_mcp_tool_to_blaxel_tool

        def traced_convert_mcp_tool_to_blaxel_tool(websocket_client, tool):
            Tool = orig_convert(websocket_client, tool)
            orig_coroutine = Tool.coroutine
            orig_sync_coroutine = Tool.sync_coroutine

            async def traced_coroutine(*args, **kwargs):
                span_attributes = {
                    "tool.name": tool.name,
                    "tool.args": json.dumps(kwargs),
                    "tool.server": websocket_client._url,
                    "tool.server_name": websocket_client.name,
                    "span.type": "tool.call",
                    **SpanManager.get_default_attributes(),
                }
                logger.info(span_attributes)
                with tracer.start_span("blaxel-tool-call", attributes=span_attributes):
                    return await orig_coroutine(*args, **kwargs)

            def traced_sync_coroutine(*args, **kwargs):
                span_attributes = {
                    "tool.name": tool.name,
                    "tool.args": json.dumps(kwargs),
                    "tool.server": websocket_client._url,
                    "tool.server_name": websocket_client.name,
                    "span.type": "tool.call",
                    **SpanManager.get_default_attributes(),
                }
                with tracer.start_span("blaxel-tool-call", attributes=span_attributes):
                    return orig_sync_coroutine(*args, **kwargs)

            Tool.coroutine = traced_coroutine
            Tool.sync_coroutine = traced_sync_coroutine
            return Tool

        blaxel.core.tools.convert_mcp_tool_to_blaxel_tool = traced_convert_mcp_tool_to_blaxel_tool



from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.flavor import Flavor


T = TypeVar("T", bound="LocationResponse")


@_attrs_define
class LocationResponse:
    """Location availability for policies

    Attributes:
        continent (Union[Unset, str]): Continent of the location
        country (Union[Unset, str]): Country of the location
        flavors (Union[Unset, list['Flavor']]): Hardware flavors available in the location
        location (Union[Unset, str]): Name of the location
        status (Union[Unset, str]): Status of the location
    """

    continent: Union[Unset, str] = UNSET
    country: Union[Unset, str] = UNSET
    flavors: Union[Unset, list["Flavor"]] = UNSET
    location: Union[Unset, str] = UNSET
    status: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        continent = self.continent

        country = self.country

        flavors: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.flavors, Unset):
            flavors = []
            for flavors_item_data in self.flavors:
                if type(flavors_item_data) is dict:
                    flavors_item = flavors_item_data
                else:
                    flavors_item = flavors_item_data.to_dict()
                flavors.append(flavors_item)

        location = self.location

        status = self.status

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if continent is not UNSET:
            field_dict["continent"] = continent
        if country is not UNSET:
            field_dict["country"] = country
        if flavors is not UNSET:
            field_dict["flavors"] = flavors
        if location is not UNSET:
            field_dict["location"] = location
        if status is not UNSET:
            field_dict["status"] = status

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.flavor import Flavor

        if not src_dict:
            return None
        d = src_dict.copy()
        continent = d.pop("continent", UNSET)

        country = d.pop("country", UNSET)

        flavors = []
        _flavors = d.pop("flavors", UNSET)
        for flavors_item_data in _flavors or []:
            flavors_item = Flavor.from_dict(flavors_item_data)

            flavors.append(flavors_item)

        location = d.pop("location", UNSET)

        status = d.pop("status", UNSET)

        location_response = cls(
            continent=continent,
            country=country,
            flavors=flavors,
            location=location,
            status=status,
        )

        location_response.additional_properties = d
        return location_response

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from logging import getLogger

from google.adk.models.lite_llm import LiteLlm, LiteLLMClient

from blaxel.core import bl_model as bl_model_core
from blaxel.core import settings

logger = getLogger(__name__)


class AuthenticatedLiteLLMClient(LiteLLMClient):
    """Provides acompletion method (for better testability)."""

    async def acompletion(self, model, messages, tools, **kwargs):
        """Asynchronously calls acompletion.

        Args:
          model: The model name.
          messages: The messages to send to the model.
          tools: The tools to use for the model.
          **kwargs: Additional arguments to pass to acompletion.

        Returns:
          The model response as a message.
        """
        kwargs["extra_headers"] = settings.auth.get_headers()
        return await super().acompletion(
            model=model,
            messages=messages,
            tools=tools,
            **kwargs,
        )

    def completion(self, model, messages, tools, stream=False, **kwargs):
        """Synchronously calls completion. This is used for streaming only.

        Args:
          model: The model to use.
          messages: The messages to send.
          tools: The tools to use for the model.
          stream: Whether to stream the response.
          **kwargs: Additional arguments to pass to completion.

        Returns:
          The response from the model.
        """
        kwargs["extra_headers"] = settings.auth.get_headers()
        return super().completion(
            model=model,
            messages=messages,
            tools=tools,
            stream=stream,
            **kwargs,
        )


async def get_google_adk_model(url: str, type: str, model: str, **kwargs):
    print("HEER TWO")
    llm_client = AuthenticatedLiteLLMClient()
    if type == "mistral":
        return LiteLlm(
            model=f"mistral/{model}",
            api_key="replaced",
            api_base=f"{url}/v1",
            llm_client=llm_client,
            **kwargs,
        )
    elif type == "cohere":
        return LiteLlm(
            model=f"cohere/{model}",
            api_base=f"{url}/v2/chat",
            llm_client=llm_client,
            **kwargs,
        )
    elif type == "xai":
        return LiteLlm(
            model=f"xai/{model}",
            api_key="replaced",
            api_base=f"{url}/v1",
            llm_client=llm_client,
            **kwargs,
        )
    elif type == "deepseek":
        return LiteLlm(
            model=f"deepseek/{model}",
            api_key="replaced",
            api_base=f"{url}/v1",
            llm_client=llm_client,
            **kwargs,
        )
    elif type == "anthropic":
        return LiteLlm(
            model=f"anthropic/{model}",
            api_base=url,
            llm_client=llm_client,
            **kwargs,
        )
    elif type == "gemini":
        return LiteLlm(
            model=f"gemini/{model}",
            api_base=f"{url}/v1beta/models/{model}",
            llm_client=llm_client,
            **kwargs,
        )
    elif type == "cerebras":
        return LiteLlm(
            model=f"cerebras/{model}",
            api_key="replaced",
            api_base=f"{url}/v1",
            llm_client=llm_client,
            **kwargs,
        )
    else:
        if type != "openai":
            logger.warning(f"Model {model} is not supported by Google ADK, defaulting to OpenAI")
        return LiteLlm(
            model=f"openai/{model}",
            api_key="replaced",
            api_base=f"{url}/v1",
            llm_client=llm_client,
            **kwargs,
        )


async def bl_model(name: str, **kwargs):
    url, type, model = await bl_model_core(name).get_parameters()
    return await get_google_adk_model(url, type, model, **kwargs)



from livekit.agents import function_tool, llm

from blaxel.core.tools import bl_tools as bl_tools_core
from blaxel.core.tools.types import Tool


def livekit_coroutine(tool: Tool):
    async def livekit_coroutine_wrapper(raw_arguments: dict[str, object]):
        result = await tool.coroutine(**raw_arguments)
        return result.model_dump_json()

    return livekit_coroutine_wrapper


def get_livekit_tools(tools: list[Tool]) -> list[llm.FunctionTool]:
    livekit_tools = []
    for tool in tools:
        livekit_tools.append(
            function_tool(
                livekit_coroutine(tool),
                raw_schema={
                    "name": tool.name,
                    "description": tool.description,
                    "parameters": tool.input_schema,
                },
            )
        )
    return livekit_tools


async def bl_tools(tools_names: list[str], **kwargs) -> list[llm.FunctionTool]:
    tools = bl_tools_core(tools_names, **kwargs)
    await tools.initialize()
    return get_livekit_tools(tools.get_tools())



import asyncio

from utils import create_or_get_sandbox

SANDBOX_NAME = "py-custom-domain"


async def main():
    sandbox = await create_or_get_sandbox(SANDBOX_NAME)
    preview = await sandbox.previews.create(
        {
            "metadata": {"name": "preview-test-public"},
            "spec": {"port": 443, "public": True, "customDomain": "prod-3.mathis.beamlit.dev"},
        }
    )
    print(preview.spec.url)


if __name__ == "__main__":
    asyncio.run(main())



import asyncio
import sys

from utils import create_or_get_sandbox

from blaxel.core.sandbox import SandboxInstance


async def main():
    """Test that validates the exec fix by creating multiple sandboxes and executing commands immediately."""
    sandbox_created = []
    try:
        i = 0
        while i < 10:
            sandbox_name = f"next-js-{i}"
            sandbox = await create_or_get_sandbox(sandbox_name=sandbox_name)

            process = await sandbox.process.exec(
                {
                    "command": "echo 'Hello, world!'",
                    "workingDir": "/blaxel",
                    "waitForCompletion": True,
                }
            )

            print(process.logs)
            sandbox_created.append(sandbox_name)
            i += 1

    except Exception as e:
        print(f"There was an error => {e}")
    finally:
        for sandbox_name in sandbox_created:
            await SandboxInstance.delete(sandbox_name)


if __name__ == "__main__":
    try:
        asyncio.run(main())
        sys.exit(0)
    except Exception as err:
        print(f"There was an error => {err}")
        sys.exit(1)



"""Blaxel - AI development platform SDK."""

from .core.common.autoload import autoload
from .core.common.env import env
from .core.common.settings import settings

autoload()

__version__ = "0.2.0"
__all__ = ["autoload", "settings", "env"]



from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="IntegrationEndpointToken")


@_attrs_define
class IntegrationEndpointToken:
    """Integration endpoint token

    Attributes:
        received (Union[Unset, str]): Integration endpoint token received
        sent (Union[Unset, str]): Integration endpoint token sent
        total (Union[Unset, str]): Integration endpoint token total
    """

    received: Union[Unset, str] = UNSET
    sent: Union[Unset, str] = UNSET
    total: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        received = self.received

        sent = self.sent

        total = self.total

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if received is not UNSET:
            field_dict["received"] = received
        if sent is not UNSET:
            field_dict["sent"] = sent
        if total is not UNSET:
            field_dict["total"] = total

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        if not src_dict:
            return None
        d = src_dict.copy()
        received = d.pop("received", UNSET)

        sent = d.pop("sent", UNSET)

        total = d.pop("total", UNSET)

        integration_endpoint_token = cls(
            received=received,
            sent=sent,
            total=total,
        )

        integration_endpoint_token.additional_properties = d
        return integration_endpoint_token

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="ResourceTrace")


@_attrs_define
class ResourceTrace:
    """Log for a resource deployment (eg. model deployment, function deployment)

    Attributes:
        duration (Union[Unset, int]): Duration in nanoseconds
        has_error (Union[Unset, bool]): Has error
        start_time (Union[Unset, str]): The timestamp of the log
        status_code (Union[Unset, int]): Status code
        trace_id (Union[Unset, str]): Trace ID of the log
    """

    duration: Union[Unset, int] = UNSET
    has_error: Union[Unset, bool] = UNSET
    start_time: Union[Unset, str] = UNSET
    status_code: Union[Unset, int] = UNSET
    trace_id: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        duration = self.duration

        has_error = self.has_error

        start_time = self.start_time

        status_code = self.status_code

        trace_id = self.trace_id

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if duration is not UNSET:
            field_dict["duration"] = duration
        if has_error is not UNSET:
            field_dict["hasError"] = has_error
        if start_time is not UNSET:
            field_dict["startTime"] = start_time
        if status_code is not UNSET:
            field_dict["statusCode"] = status_code
        if trace_id is not UNSET:
            field_dict["traceID"] = trace_id

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        if not src_dict:
            return None
        d = src_dict.copy()
        duration = d.pop("duration", UNSET)

        has_error = d.pop("hasError", UNSET)

        start_time = d.pop("startTime", UNSET)

        status_code = d.pop("statusCode", UNSET)

        trace_id = d.pop("traceID", UNSET)

        resource_trace = cls(
            duration=duration,
            has_error=has_error,
            start_time=start_time,
            status_code=status_code,
            trace_id=trace_id,
        )

        resource_trace.additional_properties = d
        return resource_trace

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import logging
from contextlib import asynccontextmanager
from typing import Any
from urllib.parse import urljoin, urlparse

import anyio
import mcp.types as types
from anyio.abc import TaskStatus
from anyio.streams.memory import MemoryObjectReceiveStream, MemoryObjectSendStream
from mcp.shared.message import SessionMessage
from websockets.asyncio.client import ClientConnection
from websockets.asyncio.client import connect as ws_connect

logger = logging.getLogger(__name__)


def remove_request_params(url: str) -> str:
    return urljoin(url, urlparse(url).path)


@asynccontextmanager
async def websocket_client(
    url: str,
    headers: dict[str, Any] | None = None,
    timeout: float = 30,
):
    """
    Client transport for WebSocket.

    The `timeout` parameter controls connection timeout.
    """
    read_stream: MemoryObjectReceiveStream[SessionMessage | Exception]
    read_stream_writer: MemoryObjectSendStream[SessionMessage | Exception]

    write_stream: MemoryObjectSendStream[SessionMessage]
    write_stream_reader: MemoryObjectReceiveStream[SessionMessage]

    read_stream_writer, read_stream = anyio.create_memory_object_stream(0)
    write_stream, write_stream_reader = anyio.create_memory_object_stream(0)

    try:
        # Wrap the task group management in a try/except GeneratorExit
        try:
            async with anyio.create_task_group() as tg:
                # Convert http(s):// to ws(s)://
                ws_url = url.replace("http://", "ws://").replace("https://", "wss://")
                logger.debug(f"Connecting to WebSocket endpoint: {remove_request_params(ws_url)}")

                # Use different parameters based on websockets version
                connection_kwargs = {"additional_headers": headers, "open_timeout": timeout}

                async with ws_connect(ws_url, **connection_kwargs) as websocket:
                    logger.debug("WebSocket connection established")

                    async def ws_reader(
                        websocket: ClientConnection,
                        task_status: TaskStatus[None] = anyio.TASK_STATUS_IGNORED,
                    ):
                        try:
                            task_status.started()
                            async for message in websocket:
                                logger.debug(f"Received WebSocket message: {message}")
                                try:
                                    parsed_message = types.JSONRPCMessage.model_validate_json(
                                        message
                                    )
                                    logger.debug(f"Received server message: {parsed_message}")
                                    await read_stream_writer.send(
                                        SessionMessage(message=parsed_message)
                                    )
                                except Exception as exc:
                                    logger.error(f"Error parsing server message: {exc}")
                                    await read_stream_writer.send(exc)
                        except Exception as exc:
                            logger.error(f"Error in ws_reader: {exc}")
                            await read_stream_writer.send(exc)
                        finally:
                            await read_stream_writer.aclose()

                    async def ws_writer(websocket: ClientConnection):
                        try:
                            async with write_stream_reader:
                                async for session_message in write_stream_reader:
                                    logger.debug(
                                        f"Sending client message: {session_message.message}"
                                    )
                                    await websocket.send(
                                        session_message.message.model_dump_json(
                                            by_alias=True,
                                            exclude_none=True,
                                        )
                                    )
                                    logger.debug("Client message sent successfully")
                        except Exception as exc:
                            logger.error(f"Error in ws_writer: {exc}")
                        finally:
                            await write_stream.aclose()

                    await tg.start(ws_reader, websocket)
                    tg.start_soon(ws_writer, websocket)

                    try:
                        yield read_stream, write_stream
                    finally:
                        # Cancel tasks when the yield returns
                        tg.cancel_scope.cancel()

        except RuntimeError:
            # Context manager is exiting via GeneratorExit, this is normal.
            # Suppress it here so it doesn't get wrapped by anyio's BaseExceptionGroup.
            # The outer finally block below will still execute for cleanup.
            pass

        # The original code had the tg.cancel_scope.cancel() inside the 'finally'
        # associated with the 'yield'. Let's ensure cancellation happens
        # reliably. It seems correctly placed already.

    finally:
        # Ensure streams are closed even if task group setup fails or GeneratorExit is caught
        await read_stream_writer.aclose()
        await write_stream.aclose()



"""
This module provides classes for handling device-based authentication,
including device login processes and bearer token management. It facilitates token refreshing
and authentication flows using device codes and bearer tokens.
"""

import base64
import json
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Dict, Generator, Optional

from httpx import Request, Response, post

from .types import BlaxelAuth, CredentialsType


@dataclass
class DeviceLogin:
    """
    A dataclass representing a device login request.

    Attributes:
        client_id (str): The client ID for the device.
        scope (str): The scope of the authentication.
    """
    client_id: str
    scope: str


@dataclass
class DeviceLoginResponse:
    """
    A dataclass representing the response from a device login request.

    Attributes:
        client_id (str): The client ID associated with the device login.
        device_code (str): The device code for authentication.
        user_code (str): The user code for completing authentication.
        expires_in (int): Time in seconds until the device code expires.
        interval (int): Polling interval in seconds.
        verification_uri (str): URI for user to verify device login.
        verification_uri_complete (str): Complete URI including the user code for verification.
    """
    client_id: str
    device_code: str
    user_code: str
    expires_in: int
    interval: int
    verification_uri: str
    verification_uri_complete: str


@dataclass
class DeviceLoginFinalizeRequest:
    """
    A dataclass representing a device login finalize request.

    Attributes:
        grant_type (str): The type of grant being requested.
        client_id (str): The client ID for finalizing the device login.
        device_code (str): The device code to finalize login.
    """
    grant_type: str
    client_id: str
    device_code: str


@dataclass
class DeviceLoginFinalizeResponse:
    access_token: str
    expires_in: int
    refresh_token: str
    token_type: str


class DeviceMode(BlaxelAuth):
    """
    DeviceMode auth that authenticates requests using a device code.
    """
    def get_headers(self) -> Dict[str, str]:
        """
        Retrieves the authentication headers containing the Bearer token and workspace information.

        Returns:
            Dict[str, str]: A dictionary of headers with Bearer token and workspace.

        Raises:
            Exception: If token refresh fails.
        """
        err = self.refresh_if_needed()
        if err:
            raise err
        return {
            "X-Blaxel-Authorization": f"Bearer {self.credentials.access_token}",
            "X-Blaxel-Workspace": self.workspace_name,
        }

    def refresh_if_needed(self) -> Optional[Exception]:
        """
        Checks if the Bearer token needs to be refreshed and performs the refresh if necessary.

        Returns:
            Optional[Exception]: An exception if refreshing fails, otherwise None.
        """
        # Need to refresh token if expires in less than 10 minutes
        parts = self.credentials.access_token.split(".")
        if len(parts) != 3:
            return Exception("Invalid JWT token format")

        try:
            claims_bytes = base64.urlsafe_b64decode(parts[1] + "=" * (-len(parts[1]) % 4))
            claims = json.loads(claims_bytes)
        except Exception as e:
            return Exception(f"Failed to decode/parse JWT claims: {str(e)}")
        exp_time = datetime.fromtimestamp(claims["exp"])
        current_time = datetime.now()
        # Refresh if token expires in less than 10 minutes
        if current_time + timedelta(minutes=10) > exp_time:
            return self.do_refresh()

        return None

    def auth_flow(self, request: Request) -> Generator[Request, Response, None]:
        """
        Processes the authentication flow by ensuring the Bearer token is valid and adding necessary headers.

        Parameters:
            request (Request): The HTTP request to authenticate.

        Yields:
            Request: The authenticated request.

        Raises:
            Exception: If token refresh fails.
        """
        err = self.refresh_if_needed()
        if err:
            return err

        request.headers["X-Blaxel-Authorization"] = f"Bearer {self.credentials.access_token}"
        request.headers["X-Blaxel-Workspace"] = self.workspace_name
        yield request

    def do_refresh(self) -> Optional[Exception]:
        """
        Performs the token refresh using the refresh token.

        Returns:
            Optional[Exception]: An exception if refreshing fails, otherwise None.
        """
        if not self.credentials.refresh_token:
            return Exception("No refresh token to refresh")

        url = f"{self.base_url}/oauth/token"
        refresh_data = {
            "grant_type": "refresh_token",
            "refresh_token": self.credentials.refresh_token,
            "device_code": self.credentials.device_code,
            "client_id": "blaxel",
        }

        try:
            response = post(url, json=refresh_data, headers={"Content-Type": "application/json"})
            response.raise_for_status()
            finalize_response = DeviceLoginFinalizeResponse(**response.json())

            if not finalize_response.refresh_token:
                finalize_response.refresh_token = self.credentials.refresh_token

            creds = CredentialsType(
                access_token=finalize_response.access_token,
                refresh_token=finalize_response.refresh_token,
                expires_in=finalize_response.expires_in,
                device_code=self.credentials.device_code,
            )

            self.credentials = creds
            return None

        except Exception as e:
            return Exception(f"Failed to refresh token: {str(e)}")

    @property
    def token(self):
        self.refresh_if_needed()
        return self.credentials.access_token



import json
from typing import Any

from agents import FunctionTool, RunContextWrapper

from blaxel.core.tools import bl_tools as bl_tools_core
from blaxel.core.tools.types import Tool


def get_openai_tool(tool: Tool) -> FunctionTool:
    async def openai_coroutine(
        _: RunContextWrapper,
        arguments: dict[str, Any],
    ) -> Any:
        result = await tool.coroutine(**json.loads(arguments))
        return result

    return FunctionTool(
        name=tool.name,
        description=tool.description,
        params_json_schema=tool.input_schema,
        on_invoke_tool=openai_coroutine,
    )


async def bl_tools(tools_names: list[str], **kwargs) -> list[FunctionTool]:
    tools = bl_tools_core(tools_names, **kwargs)
    await tools.initialize()
    return [get_openai_tool(tool) for tool in tools.get_tools()]



import asyncio
import os
from logging import getLogger

from utils import create_or_get_sandbox

from blaxel.core.sandbox import SandboxInstance

logger = getLogger(__name__)

SANDBOX_NAME = "sandbox-test-python-watch-folder"


async def main():
    """Main watch folder test function."""
    print("🚀 Starting sandbox watch folder tests...")

    try:
        # Create or get sandbox
        sandbox = await create_or_get_sandbox(SANDBOX_NAME)
        print(f"✅ Sandbox ready: {sandbox.metadata.name}")

        # Setup test directory
        user = os.environ.get("USER", "testuser")
        test_dir = f"/Users/{user}/Downloads/watch_test"

        # Clean up before test
        print("🧹 Cleaning up before test...")
        try:
            await sandbox.fs.rm(test_dir, recursive=True)
        except:
            pass

        await sandbox.fs.mkdir(test_dir)
        print(f"✅ Created test directory: {test_dir}")

        # Test 1: Basic file watch
        print("🔧 Test 1: Basic file watch...")
        watch_events = []

        def basic_watch_callback(file_event):
            watch_events.append(file_event)
            print(f"Watch event: {file_event.op} {file_event.path}/{file_event.name}")

        # Start watching
        handle = sandbox.fs.watch(test_dir, basic_watch_callback)

        # Create some files
        await sandbox.fs.write(f"{test_dir}/test1.txt", "Initial content")
        await asyncio.sleep(1)

        await sandbox.fs.write(f"{test_dir}/test2.txt", "Another file")
        await asyncio.sleep(1)

        # Modify a file
        await sandbox.fs.write(f"{test_dir}/test1.txt", "Modified content")
        await asyncio.sleep(1)

        # Delete a file
        await sandbox.fs.rm(f"{test_dir}/test2.txt")
        await asyncio.sleep(1)

        # Stop watching
        handle["close"]()

        print(f"Basic watch captured {len(watch_events)} events")
        for event in watch_events:
            print(f"  - {event.op}: {event.name}")

        # Test 2: Watch with content
        print("🔧 Test 2: Watch with content...")
        content_events = []

        def content_watch_callback(file_event):
            content_events.append(file_event)
            content_preview = file_event.content[:50] if file_event.content else "No content"
            print(f"Content watch: {file_event.op} {file_event.name} - {content_preview}")

        # Start watching with content
        content_handle = sandbox.fs.watch(test_dir, content_watch_callback, {"with_content": True})

        # Create files with specific content
        await sandbox.fs.write(
            f"{test_dir}/content1.txt", "This is the first file with some content"
        )
        await asyncio.sleep(1)

        await sandbox.fs.write(f"{test_dir}/content2.json", '{"key": "value", "number": 42}')
        await asyncio.sleep(1)

        # Modify with new content
        await sandbox.fs.write(
            f"{test_dir}/content1.txt", "This is the modified content of the first file"
        )
        await asyncio.sleep(1)

        content_handle["close"]()

        print(f"Content watch captured {len(content_events)} events")
        for event in content_events:
            content_preview = event.content[:30] if event.content else "No content"
            print(f"  - {event.op}: {event.name} ({content_preview}...)")

        # Test 3: Watch subdirectories
        print("🔧 Test 3: Watch subdirectories...")
        subdir_events = []

        def subdir_watch_callback(file_event):
            subdir_events.append(file_event)
            print(f"Subdir watch: {file_event.op} {file_event.path}/{file_event.name}")

        # Create subdirectory
        sub_dir = f"{test_dir}/subdir"
        await sandbox.fs.mkdir(sub_dir)

        # Start watching the parent directory
        subdir_handle = sandbox.fs.watch(test_dir, subdir_watch_callback)

        # Create files in subdirectory
        await sandbox.fs.write(f"{sub_dir}/sub1.txt", "File in subdirectory")
        await asyncio.sleep(1)

        await sandbox.fs.write(f"{sub_dir}/sub2.txt", "Another file in subdirectory")
        await asyncio.sleep(1)

        # Create nested subdirectory
        nested_dir = f"{sub_dir}/nested"
        await sandbox.fs.mkdir(nested_dir)
        await sandbox.fs.write(f"{nested_dir}/nested.txt", "Nested file")
        await asyncio.sleep(1)

        subdir_handle["close"]()

        print(f"Subdirectory watch captured {len(subdir_events)} events")
        for event in subdir_events:
            print(f"  - {event.op}: {event.path}/{event.name}")

        # Test 4: Multiple watchers on same directory
        print("🔧 Test 4: Multiple watchers...")
        watcher1_events = []
        watcher2_events = []

        def watcher1_callback(file_event):
            watcher1_events.append(file_event)
            print(f"Watcher 1: {file_event.op} {file_event.name}")

        def watcher2_callback(file_event):
            watcher2_events.append(file_event)
            print(f"Watcher 2: {file_event.op} {file_event.name}")

        # Start multiple watchers
        handle1 = sandbox.fs.watch(test_dir, watcher1_callback)
        handle2 = sandbox.fs.watch(test_dir, watcher2_callback, {"with_content": True})

        # Trigger events
        await sandbox.fs.write(f"{test_dir}/multi.txt", "File watched by multiple watchers")
        await asyncio.sleep(1)

        await sandbox.fs.write(f"{test_dir}/multi.txt", "Modified by multiple watchers")
        await asyncio.sleep(1)

        # Stop both watchers
        handle1["close"]()
        handle2["close"]()

        print(f"Watcher 1 captured {len(watcher1_events)} events")
        print(f"Watcher 2 captured {len(watcher2_events)} events")

        # Test 5: Watch with ignore patterns
        print("🔧 Test 5: Watch with ignore patterns...")
        ignore_events = []

        def ignore_watch_callback(file_event):
            ignore_events.append(file_event)
            print(f"Ignore watch: {file_event.op} {file_event.name}")

        # Start watching with ignore patterns
        ignore_handle = sandbox.fs.watch(
            test_dir, ignore_watch_callback, {"ignore": ["*.tmp", "*.log"]}
        )

        # Create files that should be watched
        await sandbox.fs.write(f"{test_dir}/important.txt", "This should be watched")
        await asyncio.sleep(1)

        # Create files that should be ignored
        await sandbox.fs.write(f"{test_dir}/temp.tmp", "This should be ignored")
        await asyncio.sleep(1)

        await sandbox.fs.write(f"{test_dir}/debug.log", "This should also be ignored")
        await asyncio.sleep(1)

        # Create another watched file
        await sandbox.fs.write(f"{test_dir}/another.txt", "This should also be watched")
        await asyncio.sleep(1)

        ignore_handle["close"]()

        print(f"Ignore pattern watch captured {len(ignore_events)} events")
        watched_files = [event.name for event in ignore_events]
        print(f"Watched files: {watched_files}")

        # Verify ignore patterns worked
        if "temp.tmp" not in watched_files and "debug.log" not in watched_files:
            print("✅ Ignore patterns working correctly")
        else:
            print("⚠️ Ignore patterns may not be working as expected")

        # Clean up
        print("🧹 Cleaning up test files...")
        try:
            await sandbox.fs.rm(test_dir, recursive=True)
            print("✅ Test directory removed")
        except Exception as e:
            print(f"⚠️ Failed to remove test directory: {e}")

        print("📊 Watch Test Summary:")
        print(f"  Basic watch events: {len(watch_events)}")
        print(f"  Content watch events: {len(content_events)}")
        print(f"  Subdirectory watch events: {len(subdir_events)}")
        print(f"  Multi-watcher events: W1={len(watcher1_events)}, W2={len(watcher2_events)}")
        print(f"  Ignore pattern events: {len(ignore_events)}")

        print("🎉 All watch folder tests completed!")

    except Exception as e:
        print(f"❌ Watch folder test failed with error: {e}")
        logger.exception("Watch folder test error")
        raise
    finally:
        print("🧹 Final cleanup...")
        try:
            await SandboxInstance.delete(SANDBOX_NAME)
            print("✅ Sandbox deleted")
        except Exception as e:
            print(f"⚠️ Failed to delete sandbox: {e}")


if __name__ == "__main__":
    asyncio.run(main())



import asyncio

from dotenv import load_dotenv

load_dotenv()

from logging import getLogger

from blaxel.core.jobs import bl_job

logger = getLogger(__name__)


async def main():
    """Main function for standalone execution."""
    job = bl_job("myjob")

    # Test async functionality
    print("Testing async bl_job...")
    result = await job.arun([{"name": "charlou", "age": 25}])
    logger.info(f"Async result: {result}")
    print(f"Async result: {result}")

    # Test sync functionality
    print("\nTesting sync bl_job...")
    result = job.run([{"name": "charlou", "age": 25}])
    logger.info(f"Sync result: {result}")
    print(f"Sync result: {result}")


if __name__ == "__main__":
    asyncio.run(main())



import asyncio
import os
from logging import getLogger

from utils import create_or_get_sandbox

from blaxel.core.sandbox import SandboxInstance

logger = getLogger(__name__)

SANDBOX_NAME = "sandbox-test-python-read-multiple"


async def main():
    """Main read multiple files test function."""
    print("🚀 Starting sandbox read multiple files tests...")

    try:
        # Create or get sandbox
        sandbox = await create_or_get_sandbox(SANDBOX_NAME)
        print(f"✅ Sandbox ready: {sandbox.metadata.name}")

        # Setup test directory
        user = os.environ.get("USER", "testuser")
        test_dir = f"/Users/{user}/Downloads/read_test"

        # Clean up before test
        print("🧹 Cleaning up before test...")
        try:
            await sandbox.fs.rm(test_dir, recursive=True)
        except:
            pass

        await sandbox.fs.mkdir(test_dir)
        print(f"✅ Created test directory: {test_dir}")

        # Create multiple test files with different content
        print("🔧 Creating test files...")
        test_files = {
            "file1.txt": "This is the content of file 1.\nIt has multiple lines.\nLine 3 of file 1.",
            "file2.json": '{\n  "name": "test",\n  "version": "1.0.0",\n  "description": "A test JSON file"\n}',
            "file3.md": "# Test Markdown File\n\nThis is a **markdown** file with some content.\n\n## Section 2\n\n- Item 1\n- Item 2\n- Item 3",
            "file4.py": "#!/usr/bin/env python3\n\ndef hello_world():\n    print('Hello, World!')\n\nif __name__ == '__main__':\n    hello_world()",
            "file5.csv": "Name,Age,City\nJohn,25,New York\nJane,30,San Francisco\nBob,35,Chicago",
            "empty.txt": "",
            "large.txt": "Large file content.\n" * 1000,  # Create a larger file
        }

        # Write all files
        for filename, content in test_files.items():
            file_path = f"{test_dir}/{filename}"
            await sandbox.fs.write(file_path, content)
            print(f"✅ Created {filename} ({len(content)} chars)")

        # Test 1: Read all files sequentially
        print("🔧 Test 1: Reading files sequentially...")
        sequential_results = {}
        sequential_start_time = asyncio.get_event_loop().time()

        for filename in test_files.keys():
            file_path = f"{test_dir}/{filename}"
            content = await sandbox.fs.read(file_path)
            sequential_results[filename] = content
            print(f"✅ Read {filename} ({len(content)} chars)")

        sequential_time = asyncio.get_event_loop().time() - sequential_start_time
        print(f"Sequential read time: {sequential_time:.2f} seconds")

        # Test 2: Read all files concurrently
        print("🔧 Test 2: Reading files concurrently...")
        concurrent_start_time = asyncio.get_event_loop().time()

        async def read_file(filename):
            file_path = f"{test_dir}/{filename}"
            content = await sandbox.fs.read(file_path)
            return filename, content

        # Create tasks for concurrent reading
        tasks = [read_file(filename) for filename in test_files.keys()]
        concurrent_results_list = await asyncio.gather(*tasks)
        concurrent_results = dict(concurrent_results_list)

        concurrent_time = asyncio.get_event_loop().time() - concurrent_start_time
        print(f"Concurrent read time: {concurrent_time:.2f} seconds")

        # Verify results match
        print("🔧 Verifying results...")
        for filename in test_files.keys():
            original = test_files[filename]
            sequential = sequential_results[filename]
            concurrent = concurrent_results[filename]

            if original == sequential == concurrent:
                print(f"✅ {filename}: All reads match")
            else:
                print(f"❌ {filename}: Content mismatch")
                print(f"  Original length: {len(original)}")
                print(f"  Sequential length: {len(sequential)}")
                print(f"  Concurrent length: {len(concurrent)}")

        # Test 3: Read files in subdirectories
        print("🔧 Test 3: Testing nested directory reads...")
        nested_dir = f"{test_dir}/nested"
        await sandbox.fs.mkdir(nested_dir)

        nested_files = {
            "nested/config.ini": "[section1]\nkey1=value1\nkey2=value2\n\n[section2]\nkey3=value3",
            "nested/data.xml": "<?xml version='1.0'?>\n<root>\n  <item id='1'>Value 1</item>\n  <item id='2'>Value 2</item>\n</root>",
        }

        for rel_path, content in nested_files.items():
            file_path = f"{test_dir}/{rel_path}"
            await sandbox.fs.write(file_path, content)
            print(f"✅ Created nested file {rel_path}")

        # Read nested files
        nested_results = {}
        for rel_path in nested_files.keys():
            file_path = f"{test_dir}/{rel_path}"
            content = await sandbox.fs.read(file_path)
            nested_results[rel_path] = content
            print(f"✅ Read nested {rel_path} ({len(content)} chars)")

        # Verify nested file contents
        for rel_path in nested_files.keys():
            if nested_files[rel_path] == nested_results[rel_path]:
                print(f"✅ {rel_path}: Content matches")
            else:
                print(f"❌ {rel_path}: Content mismatch")

        # Test 4: Error handling for non-existent files
        print("🔧 Test 4: Testing error handling...")
        non_existent_files = [
            f"{test_dir}/does_not_exist.txt",
            f"{test_dir}/missing/file.txt",
            "/completely/invalid/path.txt",
        ]

        for file_path in non_existent_files:
            try:
                content = await sandbox.fs.read(file_path)
                print(f"⚠️ Unexpectedly succeeded reading {file_path}")
            except Exception as e:
                print(f"✅ Expected error reading {file_path}: {type(e).__name__}")

        # Test 5: Performance comparison
        print("🔧 Test 5: Performance analysis...")
        if concurrent_time < sequential_time:
            speedup = sequential_time / concurrent_time
            print(f"✅ Concurrent reading is {speedup:.2f}x faster than sequential")
        else:
            print(
                "⚠️ Sequential reading was faster (may be due to small file sizes or network overhead)"
            )

        print("📊 Performance Summary:")
        print(f"  Files read: {len(test_files)}")
        print(f"  Sequential time: {sequential_time:.2f}s")
        print(f"  Concurrent time: {concurrent_time:.2f}s")
        print(f"  Total content size: {sum(len(content) for content in test_files.values())} chars")

        # Clean up
        print("🧹 Cleaning up test files...")
        try:
            await sandbox.fs.rm(test_dir, recursive=True)
            print("✅ Test directory removed")
        except Exception as e:
            print(f"⚠️ Failed to remove test directory: {e}")

        print("🎉 All read multiple files tests completed!")

    except Exception as e:
        print(f"❌ Read multiple files test failed with error: {e}")
        logger.exception("Read multiple files test error")
        raise
    finally:
        print("🧹 Final cleanup...")
        try:
            await SandboxInstance.delete(SANDBOX_NAME)
            print("✅ Sandbox deleted")
        except Exception as e:
            print(f"⚠️ Failed to delete sandbox: {e}")


if __name__ == "__main__":
    asyncio.run(main())



import dataclasses
import json
import logging
import os
import traceback
from contextlib import asynccontextmanager

from opentelemetry.semconv_ai import SpanAttributes


def dont_throw(func):
    # Obtain a logger specific to the function's module
    logger = logging.getLogger(func.__module__)

    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception:
            logger.debug(
                "Instrumentation failed to trace in %s, error: %s",
                func.__name__,
                traceback.format_exc(),
            )

    return wrapper


def with_tracer_wrapper(func):
    def _with_tracer(tracer):
        def wrapper(wrapped, instance, args, kwargs):
            return func(tracer, wrapped, instance, args, kwargs)

        return wrapper

    return _with_tracer


@asynccontextmanager
async def start_as_current_span_async(tracer, *args, **kwargs):
    with tracer.start_as_current_span(*args, **kwargs) as span:
        yield span

class JSONEncoder(json.JSONEncoder):
    def default(self, o):
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        elif hasattr(o, "json"):
            return o.json()
        elif hasattr(o, "to_json"):
            return o.to_json()
        return super().default(o)


def should_send_prompts():
    return (
        os.getenv("TRACELOOP_TRACE_CONTENT") or "true"
    ).lower() == "true"

@dont_throw
def process_request(span, args, kwargs):
    if should_send_prompts():
        span.set_attribute(
            SpanAttributes.TRACELOOP_ENTITY_INPUT,
            json.dumps({"args": args, "kwargs": kwargs}, cls=JSONEncoder),
        )


@dont_throw
def process_response(span, res):
    if should_send_prompts():
        span.set_attribute(
            SpanAttributes.TRACELOOP_ENTITY_OUTPUT,
            json.dumps(res, cls=JSONEncoder),
        )



import nest_asyncio

nest_asyncio.apply()

import asyncio

from dotenv import load_dotenv
from pydantic_ai import Agent, CallToolsNode
from pydantic_ai.messages import ToolCallPart

load_dotenv()

from logging import getLogger

from blaxel.core.tools import bl_tools
from blaxel.crewai.tools import bl_tools as bl_tools_crewai
from blaxel.googleadk.tools import bl_tools as bl_tools_google_adk
from blaxel.langgraph.tools import bl_tools as bl_tools_langgraph
from blaxel.llamaindex.tools import bl_tools as bl_tools_llamaindex
from blaxel.pydantic.tools import bl_tools as bl_tools_pydantic

logger = getLogger(__name__)


async def test_mcp_tools_langgraph():
    """Test bl_tools to_langgraph conversion."""
    print("Testing Langgraph tools conversion...")
    tools = await bl_tools_langgraph(["blaxel-search"])
    if len(tools) == 0:
        raise Exception("No tools found")
    result = await tools[0].ainvoke({"query": "What is the capital of France?"})
    logger.info(f"Langgraph tools result: {result}")
    print(f"Langgraph tools result: {result}")


async def test_mcp_tools_llamaindex():
    """Test llamaindex conversion."""
    print("Testing LlamaIndex tools conversion...")
    tools = await bl_tools_llamaindex(["blaxel-search"])
    if len(tools) == 0:
        raise Exception("No tools found")
    result = await tools[0].acall(query="What is the capital of France?")
    logger.info(f"LlamaIndex tools result: {result}")
    print(f"LlamaIndex tools result: {result}")


async def test_mcp_tools_crewai():
    """Test crewai conversion."""
    print("Testing CrewAI tools conversion...")
    tools = await bl_tools_crewai(["blaxel-search"])
    if len(tools) == 0:
        raise Exception("No tools found")
    result = tools[0].run(query="What is the capital of France?")
    logger.info(f"CrewAI tools result: {result}")
    print(f"CrewAI tools result: {result}")


async def test_mcp_tools_pydantic():
    """Test pydantic conversion."""
    print("Testing Pydantic tools conversion...")
    tools = await bl_tools_pydantic(["blaxel-search"])
    if len(tools) == 0:
        raise Exception("No tools found")

    agent = Agent(model="openai:gpt-4o-mini", tools=tools)
    async with agent.iter("Search what is the capital of France?") as agent_run:
        async for node in agent_run:
            # Each node represents a step in the agent's execution
            if isinstance(node, CallToolsNode):
                for part in node.model_response.parts:
                    if isinstance(part, ToolCallPart):
                        logger.info(f"Tool call: {part}")
                        print(f"Tool call: {part}")
                    else:
                        logger.info(f"Response: {part}")
                        print(f"Response: {part}")


async def test_mcp_tools_google_adk():
    """Test google adk conversion."""
    print("Testing Google ADK tools conversion...")
    tools = await bl_tools_google_adk(["blaxel-search"])
    if len(tools) == 0:
        raise Exception("No tools found")
    result = await tools[0].run_async(
        args={"query": "What is the capital of France?"}, tool_context=None
    )
    logger.info(f"Google ADK tools result: {result}")
    print(f"Google ADK tools result: {result}")


async def test_mcp_tools_blaxel():
    """Test bl_tools native blaxel functionality."""
    print("Testing native Blaxel tools...")
    tools = bl_tools(["blaxel-search"], {"test": "test"})
    await tools.initialize()
    blaxel_tools = tools.get_tools()
    if len(blaxel_tools) == 0:
        raise Exception("No tools found")
    result = await blaxel_tools[0].coroutine(query="What is the capital of France?")
    logger.info(f"Blaxel tools result: {result}")
    print(f"Blaxel tools result: {result}")
    result = await blaxel_tools[0].coroutine(query="What is the capital of Germany?")
    logger.info(f"Blaxel tools result: {result}")
    print(f"Blaxel tools result: {result}")
    await asyncio.sleep(7)
    result = await blaxel_tools[0].coroutine(query="What is the capital of USA?")
    logger.info(f"Blaxel tools result: {result}")
    print(f"Blaxel tools result: {result}")
    result = await blaxel_tools[0].coroutine(query="What is the capital of Canada?")
    logger.info(f"Blaxel tools result: {result}")
    print(f"Blaxel tools result: {result}")


async def main():
    """Main function for standalone execution."""
    # await test_mcp_tools_blaxel()
    await test_mcp_tools_langgraph()
    await test_mcp_tools_llamaindex()
    await test_mcp_tools_crewai()
    await test_mcp_tools_pydantic()
    await test_mcp_tools_google_adk()


if __name__ == "__main__":
    asyncio.run(main())



from typing import Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="ResourceLogResponse")


@_attrs_define
class ResourceLogResponse:
    """Response for a resource log

    Attributes:
        chart (Union[Unset, list[Any]]): Chart
        logs (Union[Unset, list[Any]]): Logs
        total_count (Union[Unset, int]): Total count of logs
    """

    chart: Union[Unset, list[Any]] = UNSET
    logs: Union[Unset, list[Any]] = UNSET
    total_count: Union[Unset, int] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        chart: Union[Unset, list[Any]] = UNSET
        if not isinstance(self.chart, Unset):
            chart = self.chart

        logs: Union[Unset, list[Any]] = UNSET
        if not isinstance(self.logs, Unset):
            logs = self.logs

        total_count = self.total_count

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if chart is not UNSET:
            field_dict["chart"] = chart
        if logs is not UNSET:
            field_dict["logs"] = logs
        if total_count is not UNSET:
            field_dict["totalCount"] = total_count

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        if not src_dict:
            return None
        d = src_dict.copy()
        chart = cast(list[Any], d.pop("chart", UNSET))

        logs = cast(list[Any], d.pop("logs", UNSET))

        total_count = d.pop("totalCount", UNSET)

        resource_log_response = cls(
            chart=chart,
            logs=logs,
            total_count=total_count,
        )

        resource_log_response.additional_properties = d
        return resource_log_response

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="WorkspaceUser")


@_attrs_define
class WorkspaceUser:
    """Workspace user

    Attributes:
        accepted (Union[Unset, bool]): Whether the user has accepted the workspace invitation
        email (Union[Unset, str]): Workspace user email
        email_verified (Union[Unset, bool]): Whether the user's email has been verified
        family_name (Union[Unset, str]): Workspace user family name
        given_name (Union[Unset, str]): Workspace user given name
        role (Union[Unset, str]): Workspace user role
        sub (Union[Unset, str]): Workspace user identifier
    """

    accepted: Union[Unset, bool] = UNSET
    email: Union[Unset, str] = UNSET
    email_verified: Union[Unset, bool] = UNSET
    family_name: Union[Unset, str] = UNSET
    given_name: Union[Unset, str] = UNSET
    role: Union[Unset, str] = UNSET
    sub: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        accepted = self.accepted

        email = self.email

        email_verified = self.email_verified

        family_name = self.family_name

        given_name = self.given_name

        role = self.role

        sub = self.sub

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if accepted is not UNSET:
            field_dict["accepted"] = accepted
        if email is not UNSET:
            field_dict["email"] = email
        if email_verified is not UNSET:
            field_dict["email_verified"] = email_verified
        if family_name is not UNSET:
            field_dict["family_name"] = family_name
        if given_name is not UNSET:
            field_dict["given_name"] = given_name
        if role is not UNSET:
            field_dict["role"] = role
        if sub is not UNSET:
            field_dict["sub"] = sub

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        if not src_dict:
            return None
        d = src_dict.copy()
        accepted = d.pop("accepted", UNSET)

        email = d.pop("email", UNSET)

        email_verified = d.pop("email_verified", UNSET)

        family_name = d.pop("family_name", UNSET)

        given_name = d.pop("given_name", UNSET)

        role = d.pop("role", UNSET)

        sub = d.pop("sub", UNSET)

        workspace_user = cls(
            accepted=accepted,
            email=email,
            email_verified=email_verified,
            family_name=family_name,
            given_name=given_name,
            role=role,
            sub=sub,
        )

        workspace_user.additional_properties = d
        return workspace_user

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.function_schema_not import FunctionSchemaNot
    from ..models.function_schema_properties import FunctionSchemaProperties


T = TypeVar("T", bound="FunctionSchema")


@_attrs_define
class FunctionSchema:
    """Function schema

    Attributes:
        all_of (Union[Unset, list[Any]]): List of schemas that this schema extends
        any_of (Union[Unset, list[Any]]): List of possible schemas, any of which this schema could be
        description (Union[Unset, str]): Description of the schema
        enum (Union[Unset, list[str]]): Enum values
        format_ (Union[Unset, str]): Format of the schema
        items (Union[Unset, FunctionSchema]): Function schema
        max_length (Union[Unset, float]): Maximum length for string types
        maximum (Union[Unset, float]): Maximum value for number types
        min_length (Union[Unset, float]): Minimum length for string types
        minimum (Union[Unset, float]): Minimum value for number types
        not_ (Union[Unset, FunctionSchemaNot]): Schema that this schema must not be
        one_of (Union[Unset, list[Any]]): List of schemas, one of which this schema must be
        pattern (Union[Unset, str]): Pattern for string types
        properties (Union[Unset, FunctionSchemaProperties]): Properties of the schema
        required (Union[Unset, list[str]]): Required properties of the schema
        title (Union[Unset, str]): Title of the schema
        type_ (Union[Unset, str]): Type of the schema
    """

    all_of: Union[Unset, list[Any]] = UNSET
    any_of: Union[Unset, list[Any]] = UNSET
    description: Union[Unset, str] = UNSET
    enum: Union[Unset, list[str]] = UNSET
    format_: Union[Unset, str] = UNSET
    items: Union[Unset, "FunctionSchema"] = UNSET
    max_length: Union[Unset, float] = UNSET
    maximum: Union[Unset, float] = UNSET
    min_length: Union[Unset, float] = UNSET
    minimum: Union[Unset, float] = UNSET
    not_: Union[Unset, "FunctionSchemaNot"] = UNSET
    one_of: Union[Unset, list[Any]] = UNSET
    pattern: Union[Unset, str] = UNSET
    properties: Union[Unset, "FunctionSchemaProperties"] = UNSET
    required: Union[Unset, list[str]] = UNSET
    title: Union[Unset, str] = UNSET
    type_: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        all_of: Union[Unset, list[Any]] = UNSET
        if not isinstance(self.all_of, Unset):
            all_of = self.all_of

        any_of: Union[Unset, list[Any]] = UNSET
        if not isinstance(self.any_of, Unset):
            any_of = self.any_of

        description = self.description

        enum: Union[Unset, list[str]] = UNSET
        if not isinstance(self.enum, Unset):
            enum = self.enum

        format_ = self.format_

        items: Union[Unset, dict[str, Any]] = UNSET
        if self.items and not isinstance(self.items, Unset) and not isinstance(self.items, dict):
            items = self.items.to_dict()
        elif self.items and isinstance(self.items, dict):
            items = self.items

        max_length = self.max_length

        maximum = self.maximum

        min_length = self.min_length

        minimum = self.minimum

        not_: Union[Unset, dict[str, Any]] = UNSET
        if self.not_ and not isinstance(self.not_, Unset) and not isinstance(self.not_, dict):
            not_ = self.not_.to_dict()
        elif self.not_ and isinstance(self.not_, dict):
            not_ = self.not_

        one_of: Union[Unset, list[Any]] = UNSET
        if not isinstance(self.one_of, Unset):
            one_of = self.one_of

        pattern = self.pattern

        properties: Union[Unset, dict[str, Any]] = UNSET
        if self.properties and not isinstance(self.properties, Unset) and not isinstance(self.properties, dict):
            properties = self.properties.to_dict()
        elif self.properties and isinstance(self.properties, dict):
            properties = self.properties

        required: Union[Unset, list[str]] = UNSET
        if not isinstance(self.required, Unset):
            required = self.required

        title = self.title

        type_ = self.type_

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if all_of is not UNSET:
            field_dict["allOf"] = all_of
        if any_of is not UNSET:
            field_dict["anyOf"] = any_of
        if description is not UNSET:
            field_dict["description"] = description
        if enum is not UNSET:
            field_dict["enum"] = enum
        if format_ is not UNSET:
            field_dict["format"] = format_
        if items is not UNSET:
            field_dict["items"] = items
        if max_length is not UNSET:
            field_dict["maxLength"] = max_length
        if maximum is not UNSET:
            field_dict["maximum"] = maximum
        if min_length is not UNSET:
            field_dict["minLength"] = min_length
        if minimum is not UNSET:
            field_dict["minimum"] = minimum
        if not_ is not UNSET:
            field_dict["not"] = not_
        if one_of is not UNSET:
            field_dict["oneOf"] = one_of
        if pattern is not UNSET:
            field_dict["pattern"] = pattern
        if properties is not UNSET:
            field_dict["properties"] = properties
        if required is not UNSET:
            field_dict["required"] = required
        if title is not UNSET:
            field_dict["title"] = title
        if type_ is not UNSET:
            field_dict["type"] = type_

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.function_schema_not import FunctionSchemaNot
        from ..models.function_schema_properties import FunctionSchemaProperties

        if not src_dict:
            return None
        d = src_dict.copy()
        all_of = cast(list[Any], d.pop("allOf", UNSET))

        any_of = cast(list[Any], d.pop("anyOf", UNSET))

        description = d.pop("description", UNSET)

        enum = cast(list[str], d.pop("enum", UNSET))

        format_ = d.pop("format", UNSET)

        _items = d.pop("items", UNSET)
        items: Union[Unset, FunctionSchema]
        if isinstance(_items, Unset):
            items = UNSET
        else:
            items = FunctionSchema.from_dict(_items)

        max_length = d.pop("maxLength", UNSET)

        maximum = d.pop("maximum", UNSET)

        min_length = d.pop("minLength", UNSET)

        minimum = d.pop("minimum", UNSET)

        _not_ = d.pop("not", UNSET)
        not_: Union[Unset, FunctionSchemaNot]
        if isinstance(_not_, Unset):
            not_ = UNSET
        else:
            not_ = FunctionSchemaNot.from_dict(_not_)

        one_of = cast(list[Any], d.pop("oneOf", UNSET))

        pattern = d.pop("pattern", UNSET)

        _properties = d.pop("properties", UNSET)
        properties: Union[Unset, FunctionSchemaProperties]
        if isinstance(_properties, Unset):
            properties = UNSET
        else:
            properties = FunctionSchemaProperties.from_dict(_properties)

        required = cast(list[str], d.pop("required", UNSET))

        title = d.pop("title", UNSET)

        type_ = d.pop("type", UNSET)

        function_schema = cls(
            all_of=all_of,
            any_of=any_of,
            description=description,
            enum=enum,
            format_=format_,
            items=items,
            max_length=max_length,
            maximum=maximum,
            min_length=min_length,
            minimum=minimum,
            not_=not_,
            one_of=one_of,
            pattern=pattern,
            properties=properties,
            required=required,
            title=title,
            type_=type_,
        )

        function_schema.additional_properties = d
        return function_schema

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""Blaxel telemetry module."""

from .exporters import *  # noqa: F403, F401
from .instrumentation import *  # noqa: F403, F401
from .log import *  # noqa: F403, F401
from .manager import *  # noqa: F403, F401



from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.preview_spec_request_headers import PreviewSpecRequestHeaders
    from ..models.preview_spec_response_headers import PreviewSpecResponseHeaders


T = TypeVar("T", bound="PreviewSpec")


@_attrs_define
class PreviewSpec:
    """Preview of a Resource

    Attributes:
        custom_domain (Union[Unset, str]): Custom domain bound to this preview
        port (Union[Unset, int]): Port of the preview
        prefix_url (Union[Unset, str]): Prefix URL
        public (Union[Unset, bool]): Whether the preview is public
        request_headers (Union[Unset, PreviewSpecRequestHeaders]): Those headers will be set in all requests to your
            preview. This is especially useful to set the Authorization header.
        response_headers (Union[Unset, PreviewSpecResponseHeaders]): Those headers will be set in all responses of your
            preview. This is especially useful to set the CORS headers.
        url (Union[Unset, str]): URL of the preview
    """

    custom_domain: Union[Unset, str] = UNSET
    port: Union[Unset, int] = UNSET
    prefix_url: Union[Unset, str] = UNSET
    public: Union[Unset, bool] = UNSET
    request_headers: Union[Unset, "PreviewSpecRequestHeaders"] = UNSET
    response_headers: Union[Unset, "PreviewSpecResponseHeaders"] = UNSET
    url: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        custom_domain = self.custom_domain

        port = self.port

        prefix_url = self.prefix_url

        public = self.public

        request_headers: Union[Unset, dict[str, Any]] = UNSET
        if (
            self.request_headers
            and not isinstance(self.request_headers, Unset)
            and not isinstance(self.request_headers, dict)
        ):
            request_headers = self.request_headers.to_dict()
        elif self.request_headers and isinstance(self.request_headers, dict):
            request_headers = self.request_headers

        response_headers: Union[Unset, dict[str, Any]] = UNSET
        if (
            self.response_headers
            and not isinstance(self.response_headers, Unset)
            and not isinstance(self.response_headers, dict)
        ):
            response_headers = self.response_headers.to_dict()
        elif self.response_headers and isinstance(self.response_headers, dict):
            response_headers = self.response_headers

        url = self.url

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if custom_domain is not UNSET:
            field_dict["customDomain"] = custom_domain
        if port is not UNSET:
            field_dict["port"] = port
        if prefix_url is not UNSET:
            field_dict["prefixUrl"] = prefix_url
        if public is not UNSET:
            field_dict["public"] = public
        if request_headers is not UNSET:
            field_dict["requestHeaders"] = request_headers
        if response_headers is not UNSET:
            field_dict["responseHeaders"] = response_headers
        if url is not UNSET:
            field_dict["url"] = url

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.preview_spec_request_headers import PreviewSpecRequestHeaders
        from ..models.preview_spec_response_headers import PreviewSpecResponseHeaders

        if not src_dict:
            return None
        d = src_dict.copy()
        custom_domain = d.pop("customDomain", UNSET)

        port = d.pop("port", UNSET)

        prefix_url = d.pop("prefixUrl", UNSET)

        public = d.pop("public", UNSET)

        _request_headers = d.pop("requestHeaders", UNSET)
        request_headers: Union[Unset, PreviewSpecRequestHeaders]
        if isinstance(_request_headers, Unset):
            request_headers = UNSET
        else:
            request_headers = PreviewSpecRequestHeaders.from_dict(_request_headers)

        _response_headers = d.pop("responseHeaders", UNSET)
        response_headers: Union[Unset, PreviewSpecResponseHeaders]
        if isinstance(_response_headers, Unset):
            response_headers = UNSET
        else:
            response_headers = PreviewSpecResponseHeaders.from_dict(_response_headers)

        url = d.pop("url", UNSET)

        preview_spec = cls(
            custom_domain=custom_domain,
            port=port,
            prefix_url=prefix_url,
            public=public,
            request_headers=request_headers,
            response_headers=response_headers,
            url=url,
        )

        preview_spec.additional_properties = d
        return preview_spec

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import asyncio
import logging

from blaxel.core.sandbox import SandboxInstance
from blaxel.core.sandbox.types import SandboxUpdateMetadata

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
)


async def main():
    try:
        print("Test 1: Create default sandbox...")
        sandbox = await SandboxInstance.create()

        print("Metadata:")
        print(f"  labels: {sandbox.metadata.labels if sandbox.metadata else None}")
        print(f"  displayName: {sandbox.metadata.display_name if sandbox.metadata else None}")
        print(f"  name: {sandbox.metadata.name if sandbox.metadata else None}")
        print(f"  createdAt: {sandbox.metadata.created_at if sandbox.metadata else None}")
        print(f"  updatedAt: {sandbox.metadata.updated_at if sandbox.metadata else None}")
        print(f"  createdBy: {sandbox.metadata.created_by if sandbox.metadata else None}")
        print(f"  updatedBy: {sandbox.metadata.updated_by if sandbox.metadata else None}")

        print("\nTest 2: Update metadata...")

        sandbox = await SandboxInstance.update_metadata(
            sandbox.metadata.name,
            SandboxUpdateMetadata(labels={"test": "test"}, display_name="test"),
        )

        print("Updated Metadata:")
        print(f"  labels: {sandbox.metadata.labels if sandbox.metadata else None}")
        print(f"  displayName: {sandbox.metadata.display_name if sandbox.metadata else None}")
        print(f"  name: {sandbox.metadata.name if sandbox.metadata else None}")
        print(f"  createdAt: {sandbox.metadata.created_at if sandbox.metadata else None}")
        print(f"  updatedAt: {sandbox.metadata.updated_at if sandbox.metadata else None}")
        print(f"  createdBy: {sandbox.metadata.created_by if sandbox.metadata else None}")
        print(f"  updatedBy: {sandbox.metadata.updated_by if sandbox.metadata else None}")

        await sandbox.fs.ls("/asdasdas")
        print("\nSandbox filesystem is accessible after metadata update ✅")

    except Exception as e:
        print(f"❌ There was an error => {e}")
        import traceback

        traceback.print_exc()
        exit(1)


if __name__ == "__main__":
    asyncio.run(main())
